# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting utilities for gwdetchar.lasso
"""

import os
import re
import atexit
import shutil
import tempfile
import warnings

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Alex Macedo, Jeff Bidler, Oli Patane, Marissa Walker, ' \
              'Josh Smith'


# -- plotting utilities -------------------------------------------------------

def configure_mpl_tex():
    """Configure Matplotlib with LaTeX when using multiprocessing
    """
    import matplotlib
    matplotlib.use('agg')

    from matplotlib import texmanager

    mpldir = tempfile.mkdtemp()
    atexit.register(shutil.rmtree, mpldir)
    umask = os.umask(0)
    os.umask(umask)
    os.chmod(mpldir, 0o777 & ~umask)
    os.environ['HOME'] = mpldir
    os.environ['MPLCONFIGDIR'] = mpldir

    class TexManager(texmanager.TexManager):
        texcache = os.path.join(mpldir, 'tex.cache')

    texmanager.TexManager = TexManager
    matplotlib.rcParams['ps.useafm'] = True
    matplotlib.rcParams['pdf.use14corefonts'] = True
    matplotlib.rcParams['text.usetex'] = True


def save_figure(fig, pngfile, **kwargs):
    """Save a figure
    """
    try:
        fig.save(pngfile, **kwargs)
    except (RuntimeError, IOError, IndexError):
        try:
            fig.save(pngfile, **kwargs)
        except (RuntimeError, IOError, IndexError) as e:
            warnings.warn('Error saving {0}: {1}'.format(pngfile, str(e)))
            return
    fig.close()
    return pngfile


def make_spectrum_plot(unfiltered_spectrum, filtered_spectrum, channel_name,
                       x_min, x_max, y_min, y_max, x_label, y_label, title,
                       filename):
    """Generate and return a spectrum plot
    """
    fig = unfiltered_spectrum.plot(label='Unfiltered', color='#ee0000')
    ax = fig.gca()
    ax.plot(filtered_spectrum, label='Filtered', color='#92ddc8')
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)
    ax.set_xlabel('Frequency [Hz]', fontsize='small')
    ax.set_ylabel(y_label, fontsize='small')
    ax.set_title(title, fontsize='small')
    ax.legend(loc='best', fontsize='small')
    spectrum_plot = save_figure(
        fig, filename, bbox_inches='tight')
    fig.close()
    return spectrum_plot


def make_spectrum_plots(start, end, flower, fupper, channel_name,
                        unfiltered_spectrum, filtered_spectrum):
    """Generate and return both spectrum plots
    """
    re_delim = re.compile(r'[:_-]')

    # Attributes shared between zoomed out and zoomed in plots
    channelstub = re_delim.sub('_', str(channel_name)).replace('_', '-', 1)
    gpsstub = '%d-%d' % (start, end-start)
    x_label = 'Frequency [Hz]'
    if ':GDS-CALIB_STRAIN' in channel_name:
        y_label = ('GW amplitude spectral density '
                   '[strain/$\\sqrt{\\mathrm{Hz}}]$')
    else:
        y_label = 'Primary channel units'
    y_min = 0.1 * filtered_spectrum.min().value
    y_max = 5 * filtered_spectrum.max().value

    # Atttributes specific to one of the two spectrum plots
    zoomed_out_x_min = 10
    zoomed_out_x_max = 2000
    zoomed_out_title = '%s band-passed spectrum (zoomed out)' % channelstub
    zoomed_out_spectrum_plot_name = ('%s-_BAND_PASSED_SPECTRUM_ZOOM_OUT-%s.png'
                                     % (channelstub, gpsstub))

    zoomed_in_x_min = flower - ((fupper - flower) * 0.1)
    zoomed_in_x_max = fupper + ((fupper - flower) * 0.1)
    zoomed_in_title = '%s band-passed spectrum (zoomed in)' % channelstub
    zoomed_in_spectrum_plot_name = ('%s-_BAND_PASSED_SPECTRUM_ZOOM_IN-%s.png'
                                    % (channelstub, gpsstub))

    # Generate both spectrum plots
    spectrum_plot_zoomed_out = make_spectrum_plot(
        unfiltered_spectrum, filtered_spectrum, channel_name,
        zoomed_out_x_min, zoomed_out_x_max, y_min, y_max, x_label, y_label,
        zoomed_out_title, zoomed_out_spectrum_plot_name)

    spectrum_plot_zoomed_in = make_spectrum_plot(
        unfiltered_spectrum, filtered_spectrum, channel_name,
        zoomed_in_x_min, zoomed_in_x_max, y_min, y_max, x_label, y_label,
        zoomed_in_title, zoomed_in_spectrum_plot_name)

    return (spectrum_plot_zoomed_out, spectrum_plot_zoomed_in)



# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Core utilities for implementing omega scans
"""

from scipy.signal import butter

from gwpy.segments import Segment
from gwpy.signal.qtransform import q_scan

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- basic utilities ----------------------------------------------------------

def highpass(series, f_low, order=12, analog=False, ftype='sos'):
    """High-pass a `TimeSeries` with a Butterworth filter

    Parameters
    ----------
    series : `~gwpy.timeseries.TimeSeries`
        the `TimeSeries` data to high-pass filter

    f_low : `float`
        lower cutoff frequency (Hz) of the filter

    order : `int`, optional
        number of taps in the filter, default: 12

    analog : `bool`, optional
        when True, return an analog filter, otherwise a digital filter is
        returned, default: False

    ftype : `str`, optional
        type of filter: numerator/denominator (`'ba'`), pole-zero (`'zpk'`), or
        second-order sections (`'sos'`), default: `'sos'`

    Returns
    -------
    hpseries : `~gwpy.timeseries.TimeSeries`
        the high-passed `TimeSeries`

    Notes
    -----
    This utility designs a Butterworth filter of order `order` with corner
    frequency `f_low / 1.5`, then applies this filter to the input.

    See Also
    --------
    scipy.signal.butter
    gwpy.timeseries.TimeSeries.filter
    """
    corner = f_low / 1.5
    fs = series.sample_rate.to('Hz').value
    hpfilt = butter(order, corner, btype='highpass', analog=analog,
                    output=ftype, fs=fs)
    hpseries = series.filter(hpfilt, filtfilt=True)
    return hpseries


def whiten(series, fftlength, overlap=None, method='median', window='hann',
           detrend='linear'):
    """Whiten a `TimeSeries` against its own ASD

    Parameters
    ----------
    series : `~gwpy.timeseries.TimeSeries`
        the `TimeSeries` data to whiten

    fftlength : `float`
        FFT integration length (in seconds) for ASD estimation

    overlap : `float`, optional
        seconds of overlap between FFTs, defaults to half the FFT length

    method : `str`, optional
        FFT-averaging method, default: ``'median'``,

    window : `str`, `numpy.ndarray`, optional
        window function to apply to timeseries prior to FFT,
        default: ``'hann'``
        see :func:`scipy.signal.get_window` for details on acceptable
        formats

    detrend : `str`, optional
        type of detrending to do before FFT, default: ``'linear'``

    Returns
    -------
    wseries : `~gwpy.timeseries.TimeSeries`
        a whitened version of the input data with zero mean and unit variance

    See Also
    --------
    gwpy.timeseries.TimeSeries.whiten
    """
    # get overlap window and whiten
    if overlap is None:
        overlap = fftlength / 2
    return series.whiten(fftlength=fftlength, overlap=overlap, window=window,
                         detrend=detrend, method=method).detrend(detrend)


# -- omega scans --------------------------------------------------------------

def conditioner(xoft, fftlength, overlap=None, resample=None, f_low=None,
                **kwargs):
    """Condition some input data for an omega scan

    Parameters
    ----------
    xoft : `~gwpy.timeseries.TimeSeries`
        the `TimeSeries` data to whiten

    fftlength : `float`
        FFT integration length (in seconds) for ASD estimation

    overlap : `float`, optional
        seconds of overlap between FFTs, defaults to half the FFT length

    resample : `int`, optional
        desired sampling rate (Hz) of the output if different from the input,
        default: no resampling

    f_low : `float`, optional
        lower cutoff frequency (Hz) of the filter, default: ``None``

    **kwargs : `dict`, optional
        additional arguments to :func:`highpass`

    Returns
    -------
    wxoft : `~gwpy.timeseries.TimeSeries`
        a whitened version of the input data with zero mean and unit variance

    hpxoft : `~gwpy.timeseries.TimeSeries`
        high-passed version of the input data (returned only if `f_low` is
        not ``None``)

    xoft : `~gwpy.timeseries.TimeSeries`
        original (possibly resampled) version of the input data
    """
    if resample:
        xoft = xoft.resample(resample)
    # get whitened and high-passed data streams
    if f_low is None:
        wxoft = whiten(xoft, fftlength, overlap=overlap)
        return (wxoft, xoft)
    else:
        hpxoft = highpass(xoft, f_low, **kwargs)
        wxoft = whiten(hpxoft, fftlength, overlap=overlap)
        return (wxoft, hpxoft, xoft)


def primary(gps, length, hoft, fftlength, resample=None, f_low=None,
            **kwargs):
    """Condition the primary channel for use as a matched-filter

    Parameters
    ----------
    gps : `float`
        GPS time (seconds) of suspected transient

    length : `float`
        length (seconds) of the desired matched-filter

    hoft : `~gwpy.timeseries.TimeSeries`
        the `TimeSeries` data to whiten

    fftlength : `float`
        FFT integration length (in seconds) for ASD estimation

    resample : `int`, optional
        desired sampling rate (Hz) of the output if different from the input,
        default: no resampling

    f_low : `float`, optional
        lower cutoff frequency (Hz) of the filter, default: `None`

    **kwargs : `dict`
        additional keyword arguments to `omega.conditioner`

    Returns
    -------
    out : `~gwpy.timeseries.TimeSeries`
        the conditioned data stream
    """
    if f_low is None:
        out, _ = conditioner(hoft, fftlength, resample=resample)
    else:
        out, _, _ = conditioner(
            hoft, fftlength, resample=resample, f_low=f_low, **kwargs)
    return out.crop(gps - length/2, gps + length/2).taper()


def cross_correlate(xoft, hoft):
    """Cross-correlate two `TimeSeries` by matched-filter

    Parameters
    ----------
    xoft : `~gwpy.timeseries.TimeSeries`
        the `TimeSeries` data to analyze

    hoft : `~gwpy.timeseries.TimeSeries`
        a `TimeSeries` data to use as a matched-filter

    Returns
    -------
    out : `~gwpy.timeseries.TimeSeries`
        the output of a single phase matched-filter
    """
    # make sure series have consistent sample rates
    if hoft.sample_rate.value < xoft.sample_rate.value:
        xoft = xoft.resample(hoft.sample_rate.value)
    elif hoft.sample_rate.value > xoft.sample_rate.value:
        hoft = hoft.resample(xoft.sample_rate.value)
    out = xoft.correlate(hoft, window='hann')
    return out


def scan(gps, channel, xoft, fftlength, resample=None, fthresh=1e-10,
         search=0.5, nt=1400, nf=700, logf=True, **kwargs):
    """Scan a channel for evidence of transients

    Parameters
    ----------
    gps : `float`
        the GPS time (seconds) to scan

    channel : `OmegaChannel`
        `OmegaChannel` object corresponding to this data stream

    xoft : `~gwpy.timeseries.TimeSeries`
        the `TimeSeries` data to analyze

    fftlength : `float`
        FFT integration length (in seconds) for ASD estimation

    resample : `int`, optional
        desired sampling rate (Hz) of the output if different from the input,
        default: no resampling

    fthresh : `float`, optional
        threshold on false alarm rate (Hz) for this channel to be considered
        interesting, default: 1e-10

    search : `float`, optional
        time window (seconds) around `gps` in which to find peak energies,
        default: 0.5

    nt : `int`, optional
        number of points on the time axis of the interpolated `Spectrogram`,
        default: 1400

    nf : `int`, optional
        number of points on the frequency axis of the interpolated
        `Spectrogram`, default: 700

    logf : `bool`, optional
        boolean switch to enable (`True`) or disable (`False`) use of
        log-sampled frequencies in the output `Spectrogram`, default: `True`

    **kwargs : `dict`, optional
        additional arguments to `omega.conditioner`

    Returns
    -------
    series : `tuple`
        an ordered collection of intermediate data products from this scan,
        including: the resampled `TimeSeries`, high-passed `TimeSeries`,
        whitened `TimeSeries`, whitened `QGram`, high-passed `QGram`,
        interpolated whitened `Spectrogram`, and interpolated high-passed
        `Spectrogram`
    """
    # condition data
    wxoft, hpxoft, xoft = conditioner(
        xoft, fftlength, resample=resample, f_low=channel.frange[0], **kwargs)
    # compute whitened Q-gram
    search = Segment(gps - search/2, gps + search/2)
    qgram, far = q_scan(
        wxoft, mismatch=channel.mismatch, qrange=channel.qrange,
        frange=channel.frange, search=search)
    if (far >= fthresh) and (not channel.always_plot):
        return None  # series is insignificant
    # compute raw Q-gram
    Q = qgram.plane.q
    rqgram, _ = q_scan(
        hpxoft, mismatch=channel.mismatch, qrange=(Q, Q),
        frange=qgram.plane.frange, search=search)
    # compute interpolated spectrograms
    tres = min(channel.pranges) / nt
    fres = nf if logf else (
        channel.frange[1] - channel.frange[0]) / nf
    outseg = Segment(
        gps - max(channel.pranges)/2,
        gps + max(channel.pranges)/2,
    )
    qspec = qgram.interpolate(
        tres=tres,
        fres=fres,
        logf=logf,
        outseg=outseg,
    )
    rqspec = rqgram.interpolate(
        tres=tres,
        fres=fres,
        logf=logf,
        outseg=outseg,
    )
    return (xoft, hpxoft, wxoft, qgram, rqgram, qspec, rqspec)



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwdetchar.nagios` command-line interface
"""

import json
import os
import shutil

from random import randrange
from numpy.testing import assert_equal

from .. import __main__ as nagios_cli

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

# -- command-line flags

STATUS = randrange(4)
MESSAGE = 'This is a test message'
TIMEOUT = 3600


# -- cli tests ----------------------------------------------------------------

def test_main(tmpdir, caplog):
    outdir = str(tmpdir)
    outfile = os.path.join(outdir, "test.json")
    # test the status generator
    nagios_cli.main([
        str(STATUS),
        MESSAGE,
        '--timeout', str(TIMEOUT),
        '--output-file', outfile,
    ])
    assert 'Status written to {}'.format(outfile) in caplog.text
    assert os.path.isfile(outfile)
    # test output
    with open(outfile, 'r') as fobj:
        status = json.load(fobj)
    assert isinstance(status['created_gps'], int)
    assert isinstance(status['status_intervals'], list)
    assert_equal(
        status['status_intervals'][0],
        {'start_sec': 0,
         'txt_status': MESSAGE,
         'num_status': STATUS},
    )
    assert_equal(
        status['status_intervals'][1],
        {'start_sec': TIMEOUT,
         'txt_status': 'Process timed out',
         'num_status': 3},
    )
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.lasso`
"""

import numpy
from numpy import testing as nptest

from gwpy.timeseries import (TimeSeries, TimeSeriesDict)

from .. import core

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# global test objects

IND = 8
IND2 = 25
OUTLIER_IN = numpy.random.normal(loc=0, scale=1, size=1024)
OUTLIER_IN[IND] = 100
OUTLIER_TS = TimeSeries(OUTLIER_IN, sample_rate=1024, unit='Mpc',
                        name='X1:TEST_RANGE')

OUTLIER_IN_PF = numpy.random.normal(loc=0, scale=1, size=100)
OUTLIER_IN_PF[IND] = -100
OUTLIER_IN_PF[IND2] = -75
OUTLIER_TS_PF = TimeSeries(OUTLIER_IN_PF, sample_rate=100, unit='Mpc',
                           name='X1:TEST_RANGE')

TARGET = numpy.array([-1, 0, 1])

SERIES = TimeSeries(TARGET, sample_rate=1, epoch=-1)
DATA = numpy.array([SERIES.value]).T

TSDICT = TimeSeriesDict({
    'full': SERIES,
    'flat': TimeSeries(numpy.ones(3), sample_rate=1, epoch=0),
    'nan': TimeSeries(numpy.full(3, numpy.nan), sample_rate=1, epoch=0),
})


# -- unit tests ---------------------------------------------------------------

def test_find_outliers():
    # test for standard deviation outlier finding
    # find expected outliers using standard deviation method
    outliers = core.find_outliers(OUTLIER_TS)
    assert isinstance(outliers, numpy.ndarray)
    nptest.assert_array_equal(outliers, numpy.array([IND]))


def test_find_outliers_pf():
    # test for percentile outlier finding
    # find expected outliers using percentile range method
    outliers = core.find_outliers(OUTLIER_TS_PF, N=0.01, method='pf')
    assert isinstance(outliers, numpy.ndarray)
    nptest.assert_array_equal(outliers, numpy.array([IND]))


def test_remove_outliers():
    # strip off outliers
    core.remove_outliers(OUTLIER_TS)
    assert OUTLIER_TS[IND] - OUTLIER_TS.mean() <= 5 * OUTLIER_TS.std()


def test_remove_outliers_pf():
    # Strip off outliers
    core.remove_outliers(OUTLIER_TS_PF, N=0.01, method='pf')
    outliers = core.find_outliers(OUTLIER_TS_PF, N=0.01, method='pf')
    assert isinstance(outliers, numpy.ndarray)
    nptest.assert_array_equal(outliers, numpy.array([IND2]))


def test_fit():
    # adapted from unit tests for sklearn.linear_model
    model = core.fit(DATA, TARGET, alpha=1e-8)
    assert model.alpha == 1e-8
    nptest.assert_almost_equal(model.coef_, [1])
    nptest.assert_almost_equal(model.dual_gap_, 0)
    nptest.assert_almost_equal(model.predict([[0], [1]]), [0, 1])


def test_find_alpha():
    # find the optimal alpha parameter
    alpha = core.find_alpha(DATA, TARGET)
    assert alpha == 0.1


def test_remove_flat():
    # remove flat TimeSeries
    tsdict = core.remove_flat(TSDICT)
    assert len(tsdict.keys()) == 2
    assert 'flat' not in tsdict.keys()
    nptest.assert_array_equal(tsdict['full'].value, TSDICT['full'].value)
    nptest.assert_array_equal(tsdict['nan'].value, TSDICT['nan'].value)


def test_remove_bad():
    # remove unscalable TimeSeries
    tsdict = core.remove_bad(TSDICT)
    assert len(tsdict.keys()) == 2
    assert 'nan' not in tsdict.keys()
    nptest.assert_array_equal(tsdict['full'].value, TSDICT['full'].value)
    nptest.assert_array_equal(tsdict['flat'].value, TSDICT['flat'].value)



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2020)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwdetchar.conlog` command-line interface
"""

import os
import numpy
import pytest
import shutil

from unittest import mock

from gwpy.timeseries import (
    TimeSeries,
    TimeSeriesDict,
)

from .. import conlog

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

# -- test data

(START, END) = (10, 60)
PREVIEW = 10

START_SIGNAL = TimeSeries(
    numpy.sin(2 * numpy.pi * numpy.arange(PREVIEW + 1) / 7),
    sample_rate=1,
    epoch=START - PREVIEW,
)
END_SIGNAL = TimeSeries(
    numpy.sin(2 * numpy.pi * numpy.arange(END, END + 2) / 7),
    sample_rate=1,
    epoch=END,
)

START_DATA = TimeSeriesDict({
    "H1:AUX-CHANNEL_1.mean": START_SIGNAL,
    "H1:AUX-CHANNEL_2.mean": TimeSeries(
        numpy.ones(PREVIEW + 1),
        sample_rate=1,
        epoch=START - PREVIEW,
    ),
    "H1:AUX-CHANNEL_3.mean": TimeSeries(
        numpy.ones(PREVIEW + 1),
        sample_rate=1,
        epoch=START - PREVIEW,
    ),
    "H1:AUX-CHANNEL_4.mean": TimeSeries(
        numpy.ones(PREVIEW + 1),
        sample_rate=1,
        epoch=START - PREVIEW,
    ),
})

END_DATA = TimeSeriesDict({
    "H1:AUX-CHANNEL_1.mean": END_SIGNAL,
    "H1:AUX-CHANNEL_2.mean": TimeSeries(
        numpy.ones(2),
        sample_rate=1,
        epoch=END,
    ),
    "H1:AUX-CHANNEL_3.mean": TimeSeries(
        2 * numpy.ones(2),
        sample_rate=1,
        epoch=END,
    ),
    "H1:AUX-CHANNEL_4.mean": TimeSeries(
        -4 * numpy.ones(2),
        sample_rate=1,
        epoch=END,
    ),
})


# -- cli tests ----------------------------------------------------------------

@pytest.mark.parametrize('args', (
    [],
    ['--search', 'CHANNEL_3'],
    ['--channels'],
))
@mock.patch(
    'gwdetchar.conlog._discover_data',
    return_value=(START_DATA, END_DATA),
)
@mock.patch(
    'gwdetchar.conlog._discover_data_source',
    return_value=(['/path/to/data1.gwf'],
                  ['/path/to/data2.gwf']),
)
@mock.patch(
    'gwdetchar.conlog._get_available_channels',
    return_value=set(START_DATA.keys()),
)
def test_main(gwf, src, data, args, caplog, tmpdir):
    outdir = str(tmpdir)
    outfile = os.path.join(outdir, "changes.csv")
    # write channels file
    if '--channels' in args:
        channels = os.path.join(outdir, "channels.txt")
        with open(channels, 'w') as fobj:
            fobj.write("H1:AUX-CHANNEL_4.mean")
        args.append(channels)
    # determine command-line arguments
    args = [
        str(START),
        str(END),
        '-i', 'H1',
        '--preview', str(PREVIEW),
        '--output', outfile,
    ] + args
    # test command-line tool
    conlog.main(args)
    assert 'record a state change between {0} and {1}'.format(
        START, END) in caplog.text
    assert 'Output written to {}'.format(outfile) in caplog.text
    assert os.path.getsize(outfile)
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)


@mock.patch(
    'gwdatafind.find_urls',
    return_value=[],
)
def test_main_no_data(src):
    args = [
        str(START),
        str(END),
        '-i', 'L1',
    ]
    with pytest.raises(RuntimeError) as exc:
        conlog.main(args)
    assert str(exc.value) == 'Could not find data in the time range requested'


@mock.patch(
    'gwpy.timeseries.TimeSeriesDict.read',
    return_value=END_DATA,
)
@mock.patch(
    'gwdetchar.conlog._discover_data_source',
    return_value=(['/path/to/data1.gwf'],
                  ['/path/to/data2.gwf']),
)
@mock.patch(
    'gwpy.io.gwf.iter_channel_names',
    return_value=list(END_DATA.keys()),
)
def test_main_no_change(gwf, src, data, tmpdir):
    outdir = str(tmpdir)
    outfile = os.path.join(outdir, "changes.csv")
    # determine command-line arguments
    args = [
        str(START),
        str(END),
        '-i', 'H1',
        '--preview', '1',
        '--output', outfile,
    ]
    # test command-line tool
    conlog.main(args)
    with open(outfile, 'r') as fobj:
        table = fobj.read()
    assert table == 'channel,initial_value,final_value,difference\n'
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)



# coding=utf-8
# Copyright (C) Duncan Macleod (2018)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Lasso regression utilities
"""

from math import log

import numpy
from scipy.interpolate import UnivariateSpline

from sklearn.linear_model import Lasso

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Macedo, Jeff Bidler, Oli Patane, Marissa Walker, ' \
              'Alex Urban, Josh Smith'


# -- utilities ----------------------------------------------------------------

def find_outliers(ts, N=5, method='s'):
    """Find outliers within a `TimeSeries`

    Parameters
    ----------
    ts : `~gwpy.timeseries.TimeSeries`
        data to find outliers within

    N : `float`, optional
        if `method='s'`: number of standard deviations to consider an outlier
        if `method='pf'`: percentile range limit to consider an outlier
        default for both methods: 5
    method : `str`, optional
        outlier identification method to be used, must be `'s'` (standard
        deviation method) or `'pf'` (percentil range method)
        default: `'s'`

    Returns
    -------
    out : `ndarray`
        array indices of the input where outliers occur
    """
    if method == 'pf':
        ts = ts.value  # strip out Quantity extras
        quantile = numpy.quantile(ts, N)
        outliers = []
        for i, x in enumerate(ts):
            if x < quantile:
                outliers.append(i)
        return numpy.array(outliers)
    else:
        ts = ts.value  # strip out Quantity extras
        return numpy.nonzero(abs(ts - ts.mean()) > N*ts.std())[0]


def remove_outliers(ts, N=5, method='s'):
    """Find and remove outliers within a `TimeSeries`

    Parameters
    ----------
    ts : `~gwpy.timeseries.TimeSeries`
        data to find outliers within

    N : `float`, optional
        if `method='s'`: number of standard deviations to consider an outlier
        if `method='pf'`: percentile range limit to consider an outlier
        default for both methods: 5
    method : `str`, optional
        outlier identification method to be used, must be `'s'` (standard
        deviation method) or `'pf'` (percentil range method)
        default: `'s'`

    Notes
    -----
    This action is done in-place, with no `return` statement.
    """
    if method == 'pf':
        outliers = find_outliers(ts, N=N, method='pf')
        print("There are %d outliers in this data" % len(outliers))
        unit = ts.unit
        mask = numpy.ones(ts.size, dtype=bool)
        mask[outliers] = False
        spline = UnivariateSpline(ts.times.value[mask],
                                  ts.value[mask], s=0, k=3)
        ts[outliers] = spline(ts.times.value[outliers]) * unit
        if outliers[-1] == (len(ts) - 1):
            ts = ts[:-1]
        if outliers[0] == 0:
            ts = ts[1:]
        print('Outlier removal complete')
    else:
        outliers = find_outliers(ts, N=N, method='s')
        c = 1
        while outliers.any():
            print("-- Pass %d: removing %d outliers in %s"
                  % (c, outliers.size, ts.name))
            unit = ts.unit
            cache = outliers
            mask = numpy.ones(ts.size, dtype=bool)
            mask[outliers] = False
            spline = UnivariateSpline(ts.times.value[mask], ts.value[mask],
                                      s=0, k=3)
            ts[outliers] = spline(ts.times.value[outliers]) * unit
            outliers = find_outliers(ts, N=N, method='s')
            print("   Completed %d removal passes" % c)
            if numpy.array_equal(outliers, cache):
                print("   Outliers did not change, breaking recursion")
                break
            print("   %d outliers remain" % len(outliers))
            c += 1


def fit(data, target, alpha=None):
    """Fit some data to the target using a Lasso model

    Parameters
    ----------
    data : `numpy.ndarray`
        the data

    target : `numpy.ndarray`
        the target data

    alpha : `float`
        the Lasso alpha parameter, if `None` one will be determined using
        :func:`find_alpha`

    Returns
    -------
    model : `~sklearn.linear_model.Lasso`
        the fitted model
    """
    if alpha is None:
        alpha = find_alpha(data, target)
    model = Lasso(alpha)
    return model.fit(data, target)


def find_alpha(data, target):
    """Find the best alpha value to use for the given data
    """
    # build list of alphas
    num = 100
    alphas = numpy.logspace(-1, 0, num, endpoint=True)
    nchans = numpy.zeros(num)
    coef_path = numpy.zeros((data.shape[1], num))

    # fit each alpha
    for i, alpha in enumerate(alphas):
        model = fit(data, target, alpha=alpha)
        nchans[i] = numpy.nonzero(model.coef_)[0].size
        coef_path[:, i] = model.coef_

    # prune zeros
    nonzero = nchans.nonzero()[0]
    alphas = alphas[nonzero]
    nchans = nchans[nonzero]
    coef_path = coef_path.take(nonzero, axis=1)

    # determine best alpha
    nsamps = data.shape[0]
    mean_squared_error = numpy.mean(
        (target[:, numpy.newaxis] - numpy.dot(data, coef_path)) ** 2,
        axis=0)
    sigma2 = numpy.var(target)
    eps64 = numpy.finfo('float64').eps
    criterion = (nsamps * mean_squared_error / (sigma2 + eps64)
                 + log(nsamps) * nchans)  # Eqns. 2.15--16 in (Zou et al, 2007)
    n_best = numpy.argmin(criterion)
    return alphas[n_best]


def remove_flat(tsdict):
    """Remove flat timeseries from a `TimeSeriesDict`
    """
    outdict = tsdict.copy()
    for key in tsdict.keys():
        series = tsdict[key].value
        if series.min() == series.max():
            outdict.pop(key)
    return outdict


def remove_bad(tsdict):
    """Remove data that cannot be scaled from a `TimeSeriesDict`
    """
    outdict = tsdict.copy()
    for key in tsdict.keys():
        series = tsdict[key].value
        if numpy.isnan(series).any() or numpy.isinf(series).any():
            outdict.pop(key)
    return outdict



# coding=utf-8
# Copyright (C) LIGO Scientific Collaboration (2015-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

import multiprocessing
import os
import re
import sys

from gwpy.timeseries import TimeSeries

import numpy

from MarkupPy import markup

from astropy.table import Table

from sklearn import linear_model
from sklearn.preprocessing import scale

from pandas import set_option

from gwpy.detector import ChannelList
from gwpy.io import nds2 as io_nds2

from .. import (cli, lasso as gwlasso)
from ..io.datafind import get_data
from ..io import html as htmlio

from matplotlib import (use, rcParams)
use('Agg')

# backend-dependent imports
from matplotlib.cm import get_cmap  # noqa: E402
from gwpy.plot import Plot  # noqa: E402
from gwdetchar.lasso import plot as gwplot  # noqa: E402
from gwdetchar.plot import texify  # noqa: E402

# LaTeX use
USETEX = rcParams["text.usetex"]

# default frametypes
DEFAULT_FRAMETYPE = {
    'GDS-CALIB_STRAIN': '{ifo}_HOFT_C00',
    'DMT-SNSH_EFFECTIVE_RANGE_MPC.mean': 'SenseMonitor_hoft_{ifo}_M',
    'DMT-SNSW_EFFECTIVE_RANGE_MPC.mean': 'SenseMonitor_CAL_{ifo}_M',
}

# set up logger
PROG = ('python -m gwdetchar.lasso' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- utilities ----------------------------------------------------------------

def _descaler(iterable, *coef):
    """Linear de-scaling of a data array
    """
    return (
        [((x * primary_std * coef[0]) + primary_mean) for x in iterable]
        if coef else [((x * primary_std) + primary_mean) for x in iterable]
    )


def _generate_cluster(input_):
    """Generate cluster data for use below
    """
    if USETEX:
        gwplot.configure_mpl_tex()
    currentchan = input_[1][0]
    currentts = input_[1][5]
    current = input_[0]
    p7 = (.135, .15, .95, .9)
    plot7 = None
    plot7_list = None

    if current < len(nonzerodata):
        cluster = []
        for i, otheritem in enumerate(list(auxdata.items())):
            chan_, ts_ = otheritem
            if chan_ != currentchan:
                pcorr = numpy.corrcoef(currentts.value, ts_.value)[0, 1]
                if abs(pcorr) >= cluster_threshold:
                    stub = re_delim.sub('_', chan_).replace('_', '-', 1)
                    cluster.append([i, ts_, pcorr, chan_, stub])

        if cluster:
            # write cluster table to file
            cluster = sorted(cluster, key=lambda x: abs(x[2]),
                             reverse=True)
            clustertab = Table(data=list(zip(*cluster))[2:4],
                               names=('Pearson Coefficient', 'Channel'))
            plot7_list = '%s_CLUSTER_LIST-%s.csv' % (
                re_delim.sub('_', str(currentchan)).replace('_', '-', 1),
                gpsstub)
            clustertab.write(plot7_list, format='csv', overwrite=True)

            ncluster = min(len(cluster), max_correlated_channels)
            colors2 = [cmap(i) for i in numpy.linspace(0, 1, ncluster+1)]

            # plot
            fig = Plot(figsize=(12, 4))
            fig.subplots_adjust(*p7)
            ax = fig.gca(xscale='auto-gps')
            ax.plot(
                times, scale(currentts.value)*numpy.sign(input_[1][1]),
                label=texify(currentchan), linewidth=line_size_aux,
                color=colors[0])

            for i in range(0, ncluster):
                this = cluster[i]
                ax.plot(
                    times,
                    scale(this[1].value) * numpy.sign(input_[1][1]) *
                    numpy.sign(this[2]),
                    color=colors2[i+1],
                    linewidth=line_size_aux,
                    label=('{0}, r = {1:.2}'.format(
                        texify(cluster[i][3]), cluster[i][2])),
                )

            ax.margins(x=0)
            ax.set_ylabel('Scaled amplitude [arbitrary units]')
            ax.set_title('Highly Correlated Channels')
            ax.legend(loc='center left', bbox_to_anchor=(1.05, 0.5))
            plot7 = gwplot.save_figure(fig, '%s_CLUSTER-%s.png' % (
                re_delim.sub('_', str(currentchan))
                        .replace('_', '-', 1),
                gpsstub), bbox_inches='tight')

    with counter.get_lock():
        counter.value += 1
        pc = 100 * counter.value / len(nonzerodata)
        LOGGER.info("Completed [%d/%d] %3d%% %-50s"
                    % (counter.value, len(nonzerodata), pc,
                       '(%s)' % str(currentchan)))
        sys.stdout.flush()
    return (plot7, plot7_list)


def _process_channel(input_):
    """Handle individual channels for multiprocessing
    """
    if USETEX:
        gwplot.configure_mpl_tex()
    p4 = (.1, .1, .9, .95)
    chan = input_[1][0]
    ts = input_[1][1]
    lassocoef = nonzerocoef[chan]
    zeroed = lassocoef == 0

    if zeroed:
        plot4 = None
        plot5 = None
        plot6 = None
        pcorr = None
    else:
        plot4 = None
        plot5 = None
        plot6 = None
        if trend_type == 'minute':
            pcorr = numpy.corrcoef(ts.value, primaryts.value)[0, 1]
        else:
            pcorr = 0.0
        if abs(lassocoef) < threshold:
            with counter.get_lock():
                counter.value += 1
            pc = 100 * counter.value / len(nonzerodata)
            LOGGER.info("Completed [%d/%d] %3d%% %-50s"
                        % (counter.value, len(nonzerodata), pc,
                           '(%s)' % str(chan)))
            sys.stdout.flush()
            return (chan, lassocoef, plot4, plot5, plot6, ts)

        # create time series subplots
        fig = Plot(figsize=(12, 8))
        fig.subplots_adjust(*p4)
        ax1 = fig.add_subplot(2, 1, 1, xscale='auto-gps', epoch=start)
        ax1.plot(primaryts, label=texify(primary), color='black',
                 linewidth=line_size_primary)
        ax1.set_xlabel(None)
        ax2 = fig.add_subplot(2, 1, 2, sharex=ax1, xlim=xlim)
        ax2.plot(ts, label=texify(chan), linewidth=line_size_aux)
        if range_is_primary:
            ax1.set_ylabel('Sensitive range [Mpc]')
        else:
            ax1.set_ylabel('Primary channel units')
        ax2.set_ylabel('Channel units')
        for ax in fig.axes:
            ax.legend(loc='best')
        channelstub = re_delim.sub('_', str(chan)).replace('_', '-', 1)
        plot4 = gwplot.save_figure(
            fig, '%s_TRENDS-%s.png' % (channelstub, gpsstub),
            bbox_inches='tight')

        # create scaled, sign-corrected, and overlayed timeseries
        tsscaled = scale(ts.value)
        if lassocoef < 0:
            tsscaled = numpy.negative(tsscaled)
        fig = Plot(figsize=(12, 4))
        fig.subplots_adjust(*p1)
        ax = fig.gca(xscale='auto-gps', epoch=start, xlim=xlim)
        ax.plot(times, _descaler(target), label=texify(primary),
                color='black', linewidth=line_size_primary)
        ax.plot(times, _descaler(tsscaled), label=texify(chan),
                linewidth=line_size_aux)
        if range_is_primary:
            ax.set_ylabel('Sensitive range [Mpc]')
        else:
            ax.set_ylabel('Primary Channel Units')
        ax.legend(loc='best')
        plot5 = gwplot.save_figure(
            fig, '%s_COMPARISON-%s.png' % (channelstub, gpsstub),
            bbox_inches='tight')

        # scatter plot
        tsCopy = ts.value.reshape(-1, 1)
        primarytsCopy = primaryts.value.reshape(-1, 1)
        primaryReg = linear_model.LinearRegression()
        primaryReg.fit(tsCopy, primarytsCopy)
        primaryFit = primaryReg.predict(tsCopy)
        fig = Plot(figsize=(12, 4))
        fig.subplots_adjust(*p1)
        ax = fig.gca()
        ax.set_xlabel(texify(chan) + ' [Channel units]')
        if range_is_primary:
            ax.set_ylabel('Sensitive range [Mpc]')
        else:
            ax.set_ylabel('Primary channel units')
        y_min = min(primaryts.value)
        y_max = max(primaryts.value)
        y_range = y_max - y_min
        ax.set_ylim(y_min - (y_range * 0.1), y_max + (y_range * 0.1))
        ax.text(.9, .1, 'r = ' + str('{0:.2}'.format(pcorr)),
                verticalalignment='bottom', horizontalalignment='right',
                transform=ax.transAxes, color='black', size=20,
                bbox=dict(boxstyle='square', facecolor='white', alpha=.75,
                          edgecolor='black'))
        ax.scatter(ts.value, primaryts.value, color='red')
        ax.plot(ts.value, primaryFit, color='black')
        ax.autoscale_view(tight=False, scalex=True, scaley=True)
        plot6 = gwplot.save_figure(
            fig, '%s_SCATTER-%s.png' % (channelstub, gpsstub),
            bbox_inches='tight')

    # increment counter and print status
    with counter.get_lock():
        counter.value += 1
        pc = 100 * counter.value / len(nonzerodata)
        LOGGER.info("Completed [%d/%d] %3d%% %-50s"
                    % (counter.value, len(nonzerodata), pc,
                       '(%s)' % str(chan)))
        sys.stdout.flush()
    return (chan, lassocoef, plot4, plot5, plot6, ts)


def get_primary_ts(channel, start, end, filepath=None,
                   frametype=None, cache=None, nproc=1):
    """Retrieve primary channel timeseries
    by either reading a .gwf file or querying
    """
    if filepath is not None:
        LOGGER.info('Reading primary channel file')
        return TimeSeries.read(filepath, channel=channel, start=start, end=end,
                               format='gwf', nproc=nproc)
    else:
        LOGGER.info('Querying primary channel')
        return get_data(channel, start, end,
                        verbose='Reading primary:'.rjust(30),
                        frametype=frametype, source=cache, nproc=nproc)


# -- parse command line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
        formatter_class=cli.argparse.ArgumentDefaultsHelpFormatter,
    )
    psig = parser.add_argument_group('Signal processing options')
    lsig = parser.add_argument_group('Lasso options')

    # required arguments
    cli.add_gps_start_stop_arguments(parser)
    cli.add_ifo_option(parser)

    # optional arguments
    cli.add_nproc_option(parser, default=1)
    parser.add_argument(
        '-J',
        '--nproc-plot',
        type=int,
        default=None,
        help='number of processes to use for plot rendering, '
             'will be ignored if not using LaTeX',
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        default=os.curdir,
        type=os.path.abspath,
        help='output directory for plots',
    )
    parser.add_argument(
        '-f',
        '--channel-file',
        type=os.path.abspath,
        help='path for channel file',
    )
    parser.add_argument(
        '-T',
        '--trend-type',
        default='minute',
        choices=['second', 'minute'],
        help='type of trend for correlation',
    )
    parser.add_argument(
        '-p',
        '--primary-channel',
        default='{ifo}:DMT-SNSH_EFFECTIVE_RANGE_MPC.mean',
        help='name of primary channel to use',
    )
    # primary channel filepath argument
    parser.add_argument(
        '-pf',
        '--primary-file',
        default=None,
        help='filepath of .gwf custom primary channel if using custom channel'
    )
    parser.add_argument(
        '-P',
        '--primary-frametype',
        default=None,
        help='frametype for --primary-channel, default: guess by channel name',
    )
    parser.add_argument(
        '--primary-cache',
        default=None,
        help='cache file for --primary-channel, default: None',
    )
    parser.add_argument(
        '-O',
        '--remove-outliers',
        type=float,
        default=None,
        help='Std. dev. limit for removing outliers',
    )
    parser.add_argument(
        '-R',
        '--remove-outliers-pf',
        type=float,
        default=None,
        help='Fractional limit for removing outliers between 0 and 1',
    )
    parser.add_argument(
        '-t',
        '--threshold',
        type=float,
        default=0.0001,
        help='threshold for making a plot',
    )

    # signal processing arguments
    psig.add_argument(
        '-b',
        '--band-pass',
        type=float,
        nargs=2,
        default=None,
        metavar="FLOW FHIGH",
        help='lower and upper frequencies for bandpass on h(t)',
    )
    psig.add_argument(
        '-x',
        '--filter-padding',
        type=float,
        default=3.,
        help='amount of time (seconds) to pad data for filtering',
    )

    # lasso arguments
    lsig.add_argument(
        '-a',
        '--alpha',
        default=None,
        type=float,
        help='alpha parameter for lasso fit',
    )
    lsig.add_argument(
        '-C',
        '--no-cluster',
        action='store_true',
        default=False,
        help='do not generate clustered channel plots',
    )
    lsig.add_argument(
        '-c',
        '--cluster-coefficient',
        default=.85,
        type=float,
        help='correlation coefficient threshold for clustering',
    )
    lsig.add_argument(
        '-L',
        '--line-size-primary',
        default=1,
        type=float,
        help='line width of primary channel',
    )
    lsig.add_argument(
        '-l',
        '--line-size-aux',
        default=0.75,
        type=float,
        help='line width of auxilary channel',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the lasso command-line interface
    """
    # declare global variables
    # this is needed for multiprocessing utilities
    global auxdata, cluster_threshold, cmap, colors, counter, gpsstub
    global line_size_aux, line_size_primary, max_correlated_channels
    global nonzerocoef, nonzerodata, p1, primary, primary_mean, primary_std
    global primaryts, range_is_primary, re_delim, start, target, times
    global threshold, trend_type, xlim
    parser = create_parser()
    args = parser.parse_args(args=args)

    # check for correct input
    if args.remove_outliers_pf:
        if args.remove_outliers_pf >= 1 or args.remove_outliers_pf <= 0:
            raise ValueError('Percent outlier limit must be between 0 and 1')

    # get run params
    start = int(args.gpsstart)
    end = int(args.gpsend)
    pad = args.filter_padding

    # set pertinent global variables
    cluster_threshold = args.cluster_coefficient
    line_size_aux = args.line_size_aux
    line_size_primary = args.line_size_primary
    threshold = args.threshold
    trend_type = args.trend_type

    # let's go
    LOGGER.info('{} Lasso correlations {}-{}'.format(args.ifo, start, end))

    # get primary channel frametype
    primary = args.primary_channel.format(ifo=args.ifo)
    range_is_primary = 'EFFECTIVE_RANGE_MPC' in args.primary_channel
    if args.primary_cache is not None:
        LOGGER.info("Using custom primary cache file")
    elif args.primary_frametype is None and args.primary_file is None:
        try:
            args.primary_frametype = DEFAULT_FRAMETYPE[
                primary.split(':')[1]].format(ifo=args.ifo)
        except KeyError as exc:
            raise type(exc)("Could not determine primary channel's frametype, "
                            "please specify with --primary-frametype")

    # create output directory
    if not os.path.isdir(args.output_dir):
        os.makedirs(args.output_dir)
    os.chdir(args.output_dir)

    # multiprocessing for plots
    nprocplot = (args.nproc_plot or args.nproc) if USETEX else 1

    # bandpass primary
    if args.band_pass:
        try:
            flower, fupper = args.band_pass
        except TypeError:
            flower, fupper = None

        LOGGER.info("-- Loading primary channel data")
        bandts = get_primary_ts(filepath=args.primary_file, channel=primary,
                                start=start-pad, end=end+pad,
                                frametype=args.primary_frametype,
                                cache=args.primary_cache, nproc=args.nproc)
        if flower < 0 or fupper >= float((bandts.sample_rate/2.).value):
            raise ValueError(
                "bandpass frequency is out of range for this "
                "channel, band (Hz): {0}, sample rate: {1}".format(
                    args.band_pass, bandts.sample_rate))

        # get darm BLRMS
        LOGGER.debug("-- Filtering data")
        if trend_type == 'minute':
            stride = 60
        else:
            stride = 1
        if flower:
            darmbl = (
                bandts.highpass(flower/2., fstop=flower/4.,
                                filtfilt=False, ftype='butter')
                .notch(60, filtfilt=False)
                .bandpass(flower, fupper, fstop=[flower/2., fupper*1.5],
                          filtfilt=False, ftype='butter')
                .crop(start, end))
            darmblrms = darmbl.rms(stride)
            darmblrms.name = '%s %s-%s Hz BLRMS' % (primary, flower, fupper)
        else:
            darmbl = bandts.notch(60).crop(start, end)
            darmblrms = darmbl.rms(stride)
            darmblrms.name = '%s RMS' % primary

        primaryts = darmblrms

        bandts_asd = bandts.asd(4, 2, method='median')
        darmbl_asd = darmbl.asd(4, 2, method='median')

        spectrum_plots = gwplot.make_spectrum_plots(
            start, end, flower, fupper, primary,
            bandts_asd, darmbl_asd)
        spectrum_plot_zoomed_out = spectrum_plots[0]
        spectrum_plot_zoomed_in = spectrum_plots[1]

    else:
        # load primary channel data
        LOGGER.info("-- Loading primary channel data")
        primaryts = get_primary_ts(filepath=args.primary_file, channel=primary,
                                   start=start, end=end,
                                   frametype=args.primary_frametype,
                                   cache=args.primary_cache,
                                   nproc=args.nproc).crop(start, end)

    if args.remove_outliers:
        LOGGER.debug(
            "-- Removing outliers above %f sigma" % args.remove_outliers)
        gwlasso.remove_outliers(primaryts, args.remove_outliers)
    elif args.remove_outliers_pf:
        LOGGER.debug("-- Removing outliers in the bottom {} percent "
                     "of data".format(args.remove_outliers_pf*100))
        gwlasso.remove_outliers(
            primaryts, args.remove_outliers_pf, method='pf')
        start = int(primaryts.span[0])
        end = int(primaryts.span[1])

    primary_mean = numpy.mean(primaryts.value)
    primary_std = numpy.std(primaryts.value)

    # get aux data
    LOGGER.info("-- Loading auxiliary channel data")
    if args.channel_file is None:
        host, port = io_nds2.host_resolution_order(args.ifo)[0]
        channels = ChannelList.query_nds2('*.mean', host=host, port=port,
                                          type='m-trend')
    else:
        with open(args.channel_file, 'r') as f:
            channels = [name.rstrip('\n') for name in f]
    nchan = len(channels)
    LOGGER.debug("Identified %d channels" % nchan)

    if trend_type == 'minute':
        frametype = '%s_M' % args.ifo  # for minute trends
    else:
        frametype = '%s_T' % args.ifo  # for second trends

    # read aux channels
    auxdata = get_data(
        channels, start, end, verbose='Reading:'.rjust(30),
        frametype=frametype, nproc=args.nproc, pad=0).crop(start, end)

    # -- removes flat data to be re-introdused later

    LOGGER.info('-- Pre-processing auxiliary channel data')

    auxdata = gwlasso.remove_flat(auxdata)
    flatable = Table(data=(list(set(channels) - set(auxdata.keys())),),
                     names=('Channels',))
    LOGGER.debug('Removed {0} channels with flat data'.format(len(flatable)))
    LOGGER.debug('{0} channels remaining'.format(len(auxdata)))

    # -- remove bad data

    LOGGER.info("Removing any channels with bad data...")
    nbefore = len(auxdata)
    auxdata = gwlasso.remove_bad(auxdata)
    nafter = len(auxdata)
    LOGGER.debug('Removed {0} channels with bad data'.format(nbefore - nafter))
    LOGGER.debug('{0} channels remaining'.format(nafter))
    data = numpy.array([scale(ts.value) for ts in auxdata.values()]).T

    # -- perform lasso regression -------------------

    # create model
    LOGGER.info('-- Fitting data to target')
    target = scale(primaryts.value)
    model = gwlasso.fit(data, target, alpha=args.alpha)
    LOGGER.info('Alpha: {}'.format(model.alpha))

    # restructure results for convenience
    allresults = Table(
        data=(list(auxdata.keys()), model.coef_, numpy.abs(model.coef_)),
        names=('Channel', 'Lasso coefficient', 'rank'))
    allresults.sort('rank')
    allresults.reverse()
    useful = allresults['rank'] > 0
    allresults.remove_column('rank')
    results = allresults[useful]  # non-zero coefficient
    zeroed = allresults[numpy.invert(useful)]  # zero coefficient

    # extract data for useful channels
    nonzerodata = {name: auxdata[name] for name in results['Channel']}
    nonzerocoef = {name: coeff for name, coeff in results.as_array()}

    # print results
    LOGGER.info('Found {} channels with |Lasso coefficient| >= {}:\n\n'.format(
                len(results), threshold))
    print(results)
    print('\n\n')

    # convert to pandas
    set_option('max_colwidth', -1)
    df = results.to_pandas()
    df.index += 1

    # write results to files
    gpsstub = '%d-%d' % (start, end-start)
    resultsfile = '%s-LASSO_RESULTS-%s.csv' % (args.ifo, gpsstub)
    results.write(resultsfile, format='csv', overwrite=True)
    zerofile = '%s-ZERO_COEFFICIENT_CHANNELS-%s.csv' % (args.ifo, gpsstub)
    zeroed.write(zerofile, format='csv', overwrite=True)
    flatfile = '%s-FLAT_CHANNELS-%s.csv' % (args.ifo, gpsstub)
    flatable.write(flatfile, format='csv', overwrite=True)

    # -- generate lasso plots

    modelFit = model.predict(data)

    re_delim = re.compile(r'[:_-]')
    p1 = (.1, .15, .9, .9)  # global plot defaults for plot1, lasso model

    times = primaryts.times.value
    xlim = primaryts.span
    cmap = get_cmap('tab20')
    colors = [cmap(i) for i in numpy.linspace(0, 1, len(nonzerodata)+1)]

    plot = Plot(figsize=(12, 4))
    plot.subplots_adjust(*p1)
    ax = plot.gca(xscale='auto-gps', epoch=start, xlim=xlim)
    ax.plot(times, _descaler(target), label=texify(primary),
            color='black', linewidth=line_size_primary)
    ax.plot(times, _descaler(modelFit), label='Lasso model',
            linewidth=line_size_aux)
    if range_is_primary:
        ax.set_ylabel('Sensitive range [Mpc]')
        ax.set_title('Lasso Model of Range')
    else:
        ax.set_ylabel('Primary Channel Units')
        ax.set_title('Lasso Model of Primary Channel')
    ax.legend(loc='best')
    plot1 = gwplot.save_figure(
         plot, '%s-LASSO_MODEL-%s.png' % (args.ifo, gpsstub),
         bbox_inches='tight')

    # summed contributions
    plot = Plot(figsize=(12, 4))
    plot.subplots_adjust(*p1)
    ax = plot.gca(xscale='auto-gps', epoch=start, xlim=xlim)
    ax.plot(times, _descaler(target), label=texify(primary),
            color='black', linewidth=line_size_primary)
    summed = 0
    for i, name in enumerate(results['Channel']):
        summed += scale(nonzerodata[name].value) * nonzerocoef[name]
        if i:
            label = 'Channels 1-{0}'.format(i+1)
        else:
            label = 'Channel 1'
        ax.plot(times, _descaler(summed), label=label, color=colors[i],
                linewidth=line_size_aux)
    if range_is_primary:
        ax.set_ylabel('Sensitive range [Mpc]')
    else:
        ax.set_ylabel('Primary Channel Units')
    ax.set_title('Summations of Channel Contributions to Model')
    ax.legend(loc='center left', bbox_to_anchor=(1.05, 0.5))
    plot2 = gwplot.save_figure(
        plot, '%s-LASSO_CHANNEL_SUMMATION-%s.png' % (args.ifo, gpsstub),
        bbox_inches='tight')

    # individual contributions
    plot = Plot(figsize=(12, 4))
    plot.subplots_adjust(*p1)
    ax = plot.gca(xscale='auto-gps', epoch=start, xlim=xlim)
    ax.plot(times, _descaler(target), label=texify(primary),
            color='black', linewidth=line_size_primary)
    for i, name in enumerate(results['Channel']):
        this = _descaler(scale(nonzerodata[name].value) * nonzerocoef[name])
        if i:
            label = 'Channels 1-{0}'.format(i+1)
        else:
            label = 'Channel 1'
        ax.plot(times, this, label=texify(name), color=colors[i],
                linewidth=line_size_aux)
    if range_is_primary:
        ax.set_ylabel('Sensitive range [Mpc]')
    else:
        ax.set_ylabel('Primary Channel Units')
    ax.set_title('Individual Channel Contributions to Model')
    ax.legend(loc='center left', bbox_to_anchor=(1.05, 0.5))
    plot3 = gwplot.save_figure(
        plot, '%s-LASSO_CHANNEL_CONTRIBUTIONS-%s.png' % (args.ifo, gpsstub),
        bbox_inches='tight')

    # -- process aux channels, making plots

    LOGGER.info("-- Processing channels")
    counter = multiprocessing.Value('i', 0)

    # process channels
    pool = multiprocessing.Pool(nprocplot)
    results = pool.map(_process_channel, enumerate(list(nonzerodata.items())))
    results = sorted(results, key=lambda x: abs(x[1]), reverse=True)

    #  generate clustered time series plots
    counter = multiprocessing.Value('i', 0)
    max_correlated_channels = 20

    if args.no_cluster is False:
        LOGGER.info("-- Generating clusters")
        pool = multiprocessing.Pool(nprocplot)
        clusters = pool.map(_generate_cluster, enumerate(results))

    channelsfile = '%s-CHANNELS-%s.csv' % (args.ifo, gpsstub)
    numpy.savetxt(channelsfile, channels, delimiter=',', fmt='%s')

    # write html
    trange = '%d-%d' % (start, end)
    title = '%s Lasso Correlation: %s' % (args.ifo, trange)
    if args.band_pass:
        links = [trange] + [(s, '#%s' % s.lower()) for s in
                            ['Parameters', 'Spectra', 'Model', 'Results']]
    else:
        links = [trange] + [(s, '#%s' % s.lower()) for s in
                            ['Parameters', 'Model', 'Results']]
    (brand, class_) = htmlio.get_brand(args.ifo, 'Lasso', start)
    navbar = htmlio.navbar(links, class_=class_, brand=brand)
    page = htmlio.new_bootstrap_page(
        title='%s Lasso | %s' % (args.ifo, trange),
        navbar=navbar)
    page.h1(title, class_='pb-2 mt-3 mb-2 border-bottom')

    # -- summary table
    content = [
        ('Primary channel', markup.oneliner.code(primary)),
        ('Primary frametype', markup.oneliner.code(
            args.primary_frametype) or '-'),
        ('Primary cache file', markup.oneliner.code(
            args.primary_cache) or '-'),
        ('Outlier threshold', '%s sigma' % args.remove_outliers),
        ('Lasso coefficient threshold', str(threshold)),
        ('Cluster coefficient threshold', str(args.cluster_coefficient)),
        ('Non-zero coefficients', str(numpy.count_nonzero(model.coef_))),
        ('&alpha; (model)', '%.4f' % model.alpha)]
    if args.band_pass:
        content.insert(2, ('Primary bandpass',
                           '{0}-{1} Hz'.format(flower, fupper)))
    page.h2('Parameters', class_='mt-4 mb-4', id_='parameters')
    page.div(class_='row')
    page.div(class_='col-md-9 col-sm-12')
    page.add(htmlio.parameter_table(content, start=start, end=end))
    page.div.close()  # col-md-9 col-sm-12

    # -- download button
    files = [
        ('%s analyzed channels (CSV)' % nchan, channelsfile),
        ('%s flat channels (CSV)' % len(flatable), flatfile),
        ('%s zeroed channels (CSV)' % len(zeroed), zerofile)]
    page.div(class_='col-md-3 col-sm-12')
    page.add(htmlio.download_btn(
        files, label='Channel information',
        btnclass='btn btn-%s dropdown-toggle' % args.ifo.lower()))
    page.div.close()  # col-md-3 col-sm-12
    page.div.close()  # rowa

    # -- command-line
    page.h5('Command-line:')
    page.add(htmlio.get_command_line(about=False, prog=PROG))

    if args.band_pass:
        page.h2('Primary channel spectra', class_='mt-4', id_='spectra')
        page.div(class_='card border-light card-body shadow-sm')
        page.div(class_='row')
        page.div(class_='col-md-6')
        spectra_img1 = htmlio.FancyPlot(spectrum_plot_zoomed_out)
        page.add(htmlio.fancybox_img(spectra_img1))
        page.div.close()  # col-md-6
        page.div(class_='col-md-6')
        spectra_img2 = htmlio.FancyPlot(spectrum_plot_zoomed_in)
        page.add(htmlio.fancybox_img(spectra_img2))
        page.div.close()  # col-md-6
        page.div.close()  # row
        page.div.close()  # card border-light card-body shadow-sm

    # -- model information
    page.h2('Model information', class_='mt-4', id_='model')

    page.div(class_='card card-%s card-body shadow-sm' % args.ifo.lower())
    page.div(class_='row')
    page.div(class_='col-md-8 offset-md-2', id_='results-table')
    page.p('Below are the top {} mean minute-trend channels, ranked by '
           'Lasso correlation with the primary.'.format(df.shape[0]))
    page.add(df.to_html(
        classes=('table', 'table-sm', 'table-hover'),
        formatters={
            'Lasso coefficient': lambda x: "%.4f" % x,
            'Channel': lambda x: str(htmlio.cis_link(x.split('.')[0])),
            '__index__': lambda x: str(x)
        },
        escape=False,
        border=0).replace(' style="text-align: right;"', ''))
    page.div.close()  # col-md-10 offset-md-1
    page.div.close()  # row

    page.div(class_='row', id_='primary-lasso')
    page.div(class_='col-md-8 offset-md-2')
    img1 = htmlio.FancyPlot(plot1)
    page.add(htmlio.fancybox_img(img1))  # primary lasso plot
    page.div.close()  # col-md-8 offset-md-2
    page.div.close()  # primary-lasso

    page.div(class_='row', id_='channel-summation')
    img2 = htmlio.FancyPlot(plot2)
    page.div(class_='col-md-8 offset-md-2')
    page.add(htmlio.fancybox_img(img2))
    page.div.close()  # col-md-8 offset-md-2
    page.div.close()  # channel-summation

    page.div(class_='row', id_='channels-and-primary')
    img3 = htmlio.FancyPlot(plot3)
    page.div(class_='col-md-8 offset-md-2')
    page.add(htmlio.fancybox_img(img3))
    page.div.close()  # col-md-8 offset-md-2
    page.div.close()  # channels-and-primary

    page.div.close()  # card card-<ifo> card-body shadow-sm

    # -- results
    page.h2('Top channels', class_='mt-4', id_='results')
    page.div(id_='results')
    # for each aux channel, create information container and put plots in it
    for i, (ch, lassocoef, plot4, plot5, plot6, ts) in enumerate(results):
        # set container color/context based on lasso coefficient
        if lassocoef == 0:
            break
        elif abs(lassocoef) < threshold:
            h = '%s [lasso coefficient = %.4f] (Below threshold)' % (
                ch, lassocoef)
        else:
            h = '%s [lasso coefficient = %.4f]' % (ch, lassocoef)
        if ((lassocoef is None) or (lassocoef == 0)
                or (abs(lassocoef) < threshold)):
            card = 'card border-light mb-1 shadow-sm'
            card_header = 'card-header bg-light'
        elif abs(lassocoef) >= .5:
            card = 'card border-danger mb-1 shadow-sm'
            card_header = 'card-header text-white bg-danger'
        elif abs(lassocoef) >= .2:
            card = 'card border-warning mb-1 shadow-sm'
            card_header = 'card-header text-white bg-warning'
        else:
            card = 'card border-info mb-1 shadow-sm'
            card_header = 'card-header text-white bg-info'
        page.div(class_=card)

        # heading
        page.div(class_=card_header)
        page.a(h, class_='collapsed card-link cis-link', href='#channel%d' % i,
               **{'data-toggle': 'collapse'})
        page.div.close()  # card-header
        # body
        page.div(id_='channel%d' % i, class_='collapse',
                 **{'data-parent': '#results'})
        page.div(class_='card-body')
        if lassocoef is None:
            page.p('The amplitude data for this channel is flat (does not '
                   'change) within the chosen time period.')
        elif abs(lassocoef) < threshold:
            page.p('Lasso coefficient below the threshold of %g.'
                   % (threshold))
        else:
            for image in [plot4, plot5, plot6]:
                img = htmlio.FancyPlot(image)
                page.div(class_='row')
                page.div(class_='col-md-8 offset-md-2')
                page.add(htmlio.fancybox_img(img))
                page.div.close()  # col-md-8 offset-md-2
                page.div.close()  # row
                page.add('<hr class="row-divider">')
            if args.no_cluster is False:
                if clusters[i][0] is None:
                    page.p("<font size='3'><br />No channels were highly "
                           "correlated with this channel.</font>")
                else:
                    page.div(class_='row', id_='clusters')
                    page.div(class_='col-md-12')
                    cimg = htmlio.FancyPlot(clusters[i][0])
                    page.add(htmlio.fancybox_img(cimg))
                    page.div.close()  # col-md-12
                    page.div.close()  # clusters
                    if clusters[i][1] is not None:
                        corr_link = markup.oneliner.a(
                            'Export {} channels (CSV)'.format(
                                max_correlated_channels),
                            href=clusters[i][1], download=clusters[i][1],
                        )
                        page.button(
                            corr_link,
                            class_='btn btn-%s' % args.ifo.lower(),
                        )
        page.div.close()  # card-body
        page.div.close()  # collapse
        page.div.close()  # card
    page.div.close()  # results
    htmlio.close_page(page, 'index.html')  # save and close
    LOGGER.info("-- Process Completed")


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2020)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwdetchar.scattering.simple` command-line interface
"""

import os
import numpy
import shutil

from numpy.testing import assert_equal
from unittest import mock

from gwpy.timeseries import (
    TimeSeries,
    TimeSeriesDict,
)

from .. import simple

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

IFO = 'L1'

# -- test data

# The unit test below captures all use cases by simulating
#     (1) a missing optic channel, and
#     (2) a scattering fringe in h(t) predicted by only one optic.

DURATION = 68
FREQ = 1 / 10
SAMPLE = 4096

TIMES = numpy.arange(0, DURATION, 1 / SAMPLE)
PHASE = 42 * numpy.sin(2 * numpy.pi * FREQ * TIMES) / (2 * numpy.pi * FREQ)
SCATTER = TimeSeries(
    (numpy.sin(numpy.pi * TIMES / DURATION)
     * numpy.cos(2 * numpy.pi * PHASE)),
    sample_rate=SAMPLE,
)

HOFT = TimeSeries(
    numpy.random.normal(loc=1, scale=1.5, size=SCATTER.size),
    sample_rate=SAMPLE,
).inject(SCATTER.highpass(10))

AUX = TimeSeriesDict({
    ':'.join([IFO, chan]): TimeSeries(
        numpy.random.normal(loc=1, scale=1e-3, size=SCATTER.size),
        sample_rate=SAMPLE,
        name=':'.join([IFO, chan]),
    ).crop(4, 64) for chan in simple.MOTION_CHANNELS[1::]
})

PHASE = PHASE[4 * SAMPLE:-4 * SAMPLE]
AUX['{}:SUS-ITMX_R0_DAMP_L_IN1_DQ'.format(IFO)] += 1.064 * PHASE / 2


# -- cli tests ----------------------------------------------------------------

@mock.patch(
    'gwdetchar.scattering.simple._discover_data',
    return_value=(HOFT, AUX),
)
def test_main(data, caplog, tmpdir):
    outdir = str(tmpdir)
    args = [  # command-line arguments
        str(DURATION / 2),
        '-i', IFO,
        '--multipliers', '1',
        '--output-dir', outdir,
    ]
    # test command-line tool
    simple.main(args)
    assert 1 == caplog.text.count(
        "Skipping {}:SUS-BS_M1_DAMP_L_IN1_DQ".format(IFO))
    assert len(simple.MOTION_CHANNELS) - 2 == caplog.text.count(
        "No significant evidence of scattering found")
    assert 1 == caplog.text.count(
        "Plotting spectra and projected fringe frequencies")
    assert 1 == caplog.text.count("1 chanels plotted")
    # test output
    assert_equal(
        set(os.listdir(outdir)),
        {"{}-SUS_ITMX_R0_DAMP_L_IN1_DQ-34.0-60-comparison.png".format(IFO),
         "{}-SUS_ITMX_R0_DAMP_L_IN1_DQ-34.0-60-overlay.png".format(IFO)},
    )
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.omega.batch`
"""

import os
import pytest
import shutil
import numpy.testing as nptest

from getpass import getuser
from pycondor import Dagman
from unittest import mock

from .. import batch

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

# global test objects

FLAGS = [
    '--ifo', 'X1',
    '--frequency-scaling', 'log',
    '--colormap', 'viridis',
    '--nproc', '8',
    '--far-threshold', '3.171e-08',
    '--disable-correlation',
    '--disable-checkpoint',
    '--ignore-state-flags',
]

CONDORCMDS = [
    'accounting_group = ligo.dev.o1.detchar.user_req.omegascan',
    'accounting_group_user = {}'.format(getuser()),
    'periodic_remove = CurrentTime-EnteredCurrentStatus > 14400',
    'test',
]


# -- unit tests ---------------------------------------------------------------

def test_get_command_line_flags():
    ifo = 'X1'
    flags = batch.get_command_line_flags(
        ifo, disable_correlation=True, disable_checkpoint=True,
        ignore_state_flags=True)
    nptest.assert_array_equal(flags, FLAGS)


def test_get_condor_arguments():
    gps = 1126259462
    condorcmds = batch.get_condor_arguments(
        timeout=4, extra_commands=['test'], gps=gps)
    nptest.assert_array_equal(condorcmds, CONDORCMDS)


@mock.patch('pycondor.Dagman.submit_dag', return_value=None)
def test_generate_dag(dag, tmpdir, capsys):
    outdir = str(tmpdir)
    times = [1187008882]
    # test without submit
    dagman = batch.generate_dag(
        times, flags=FLAGS, outdir=outdir, condor_commands=CONDORCMDS)
    assert isinstance(dagman, Dagman)
    assert os.listdir(os.path.join(outdir, 'condor'))
    assert os.path.isdir(os.path.join(outdir, 'logs'))
    (out, err) = capsys.readouterr()
    assert not err
    assert out.startswith('The directory')
    assert 'Submit to condor via' in out
    # test with submit
    dagman = batch.generate_dag(
        times, flags=FLAGS, submit=True, outdir=outdir,
        condor_commands=CONDORCMDS)
    (out, err) = capsys.readouterr()
    assert not err
    assert 'condor_q' in out
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)


# -- cli tests ----------------------------------------------------------------

@pytest.mark.parametrize('args', (
    ['1126259462.4', '--ifo', 'H1'],
    ['1128678900.4', '1135136350.6', '--submit', '--ifo', 'L1'],
    ['event-times.txt', '--submit', '--monitor', '--ifo', 'Network'],
))
@mock.patch('numpy.loadtxt', return_value=[
    1167559936.6, 1180922494.5, 1185389807.3])
@mock.patch('pycondor.Dagman.submit_dag', return_value=None)
@mock.patch('subprocess.check_call', return_value=None)
def test_main(subp, dag, loadtxt, args, tmpdir):
    outdir = str(tmpdir)
    batch.main(args + ['--output-dir', outdir])
    # test output
    assert os.listdir(os.path.join(outdir, 'condor'))
    assert os.path.isdir(os.path.join(outdir, 'logs'))
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)



# coding=utf-8
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting utilities
"""

from matplotlib import rcParams

from gwpy.plot.tex import label_to_latex

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Dan Hoak <daniel.hoak@ligo.org>, ' \
              'Duncan Macleod <duncan.macleod@ligo.org>'


# -- plotting utilities -------------------------------------------------------

def texify(text):
    """Helper utility to detect when LaTeX rendering is used, and convert
    text to a LaTeX-passable representation if necessary

    Parameters
    ----------
    text : str
        text to convert to LaTeX representation

    Returns
    -------
    out : str
        either a copy or LaTeX representation of `text`

    See Also
    --------
    gwpy.plot.tex.label_to_latex
        the underlying method to convert to a LaTeX representation
    """
    if rcParams['text.usetex']:
        return label_to_latex(text)
    return text or ''


def plot_segments(flag, span, facecolor='red', edgecolor='darkred', height=0.8,
                  known={'alpha': 0.6, 'facecolor': 'lightgray',
                         'edgecolor': 'gray', 'height': 0.4}):
    """Plot the saturation segments contained within a flag
    """
    name = texify(flag.name)
    plot = flag.plot(
        figsize=[12, 2], facecolor=facecolor, edgecolor=edgecolor,
        height=height, known=known, label=' ', xlim=span, xscale='auto-gps',
        epoch=span[0], title="{} segments".format(name),
    )
    plot.subplots_adjust(bottom=0.4, top=0.8)
    return plot



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.scattering`
"""

import numpy
from numpy import testing as nptest

from gwpy.timeseries import TimeSeries
from gwpy.testing.utils import assert_segmentlist_equal

from .. import core

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# global test objects

TWOPI = 2*numpy.pi
TIMES = numpy.arange(0, 32, 1./2048)
OPTIC = TimeSeries(
    numpy.cos(TWOPI*10*TIMES), sample_rate=2048, name='X1:TEST')


# -- make sure plots run end-to-end -------------------------------------------

def test_get_fringe_frequency():
    # calculate the fringe frequency
    fringef = core.get_fringe_frequency(OPTIC, multiplier=1)
    assert str(fringef.unit) == 'Hz'
    assert fringef.sample_rate.value == OPTIC.sample_rate.value
    assert fringef.size == OPTIC.size
    nptest.assert_almost_equal(
        fringef.value.max() * (1.064 / 2) / TWOPI, 10, decimal=2)


def test_get_blrms():
    # calculate the whitened, band-limited RMS
    fringef = core.get_fringe_frequency(OPTIC, multiplier=1)
    wblrms = core.get_blrms(fringef, fhigh=20, whiten=False)
    nptest.assert_array_equal(  # BLRMS should be equivalent
        wblrms.value, fringef.bandpass(4, 20).rms(1).value)


def test_get_segments():
    # get segments from optic motion
    fringef = core.get_fringe_frequency(OPTIC, multiplier=1)
    dqflag = core.get_segments(fringef, 10)
    assert dqflag.name == fringef.name
    assert len(dqflag.active) == 640
    assert_segmentlist_equal(dqflag.known, [OPTIC.span])



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Command-line interface utilities for `gwdetchar`
"""

import argparse
import coloredlogs
import datetime
import logging
import sys

from pytz import reference

from gwpy.time import to_gps

from . import (const, __version__)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credit__ = 'Alex Urban <alexander.urban@ligo.org>'

# global variables

LEVEL_STYLES = {
    'critical': {'color': 9, 'bold': True},
    'debug': {'color': 14},
    'error': {'color': 'red'},
    'info': {'color': 10},
    'notice': {'color': 'magenta'},
    'spam': {'color': 'green', 'faint': True},
    'success': {'color': 'green', 'bold': True},
    'verbose': {'color': 'blue'},
    'warning': {'color': 13},
}

FIELD_STYLES = {
    'levelname': {'color': 39},
    'asctime': {'color': 27},
    'name': {'color': 12},
}

NOW = datetime.datetime.now()
TIMEZONE = reference.LocalTimezone().tzname(NOW)

DATEFMT = '%Y-%m-%d %H:%M:%S {}'.format(TIMEZONE)
FMT = '%(name)s %(asctime)s %(levelname)+8s: %(message)s'

# disable matplotlib logging
logging.getLogger("matplotlib").setLevel(logging.CRITICAL)


# -- logging and parsing utilities --------------------------------------------

def logger(name=__name__, level='DEBUG'):
    """Construct a logger utility for stdout/stderr messages
    """
    logger = logging.getLogger(name)
    coloredlogs.install(
        level=level, logger=logger, stream=sys.stdout, fmt=FMT,
        datefmt=DATEFMT, level_styles=LEVEL_STYLES, field_styles=FIELD_STYLES)
    logger.setLevel(level)
    return logger


def create_parser(**kwargs):
    """Create a new `argparse.ArgumentParser`
    """
    kwargs.setdefault(
        'formatter_class',
        argparse.RawDescriptionHelpFormatter,
    )
    version = kwargs.pop('version', __version__)
    parser = argparse.ArgumentParser(**kwargs)
    if version is not None:
        add_version_option(parser, version=version)
    return parser


def add_version_option(parser, version=None):
    return parser.add_argument('-V', '--version', action='version',
                               version=(version or __version__))


def add_ifo_option(parser, ifo=const.IFO, required=None):
    """Add a `-i/--ifo` option to the given parser
    """
    if required is None:
        required = const.IFO is None
    return parser.add_argument(
        '-i', '--ifo', default=const.IFO, required=required,
        help='IFO prefix for this analysis, default: %(default)s')


def add_gps_start_stop_arguments(parser, type=to_gps, **kwargs):
    """Add `gpsstart` and `gpsend` arguments to the given parser
    """
    a = parser.add_argument('gpsstart', type=to_gps,
                            help='GPS start time or datetime of analysis',
                            **kwargs)
    b = parser.add_argument('gpsend', type=to_gps,
                            help='GPS end time or datetime of analysis',
                            **kwargs)
    return (a, b)


def add_gps_start_stop_options(parser, type=to_gps, **kwargs):
    """Add `-s/--gps-start-time` and `-e/--gps-end-time` arguments
    """
    a = parser.add_argument('-s', '--gps-start-time', type=to_gps,
                            help='GPS start time or datetime of analysis',
                            **kwargs)
    b = parser.add_argument('-e', '--gps-end-time', type=to_gps,
                            help='GPS end time or datetime of analysis',
                            **kwargs)
    return (a, b)


def add_frametype_option(parser, **kwargs):
    """Add a `-f/--frametype` option to this given parser
    """
    kwargs.setdefault('help', 'the frame type name')
    return parser.add_argument('-f', '--frametype', **kwargs)


def add_nproc_option(
        parser, default=8, type=int,
        help='the number of processes to use when reading data, '
             'default: %(default)s',
        **kwargs):
    return parser.add_argument('-j', '--nproc', default=default, help=help,
                               type=type, **kwargs)



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the GW DetChar python package.
#
# gwdetchar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# gwdetchar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Test suite for `gwdetchar.const`
"""

import os

try:
    from importlib import reload
except ImportError:  # python < 3
    reload = reload

import pytest

from .. import const as _const

_DEFAULT_ENV = os.environ.copy()


@pytest.fixture
def const():
    yield _const
    reload(_const)
    os.environ.update(_DEFAULT_ENV)


@pytest.mark.parametrize('env', [
    {},
    {'IFO': 'X1'},
])
def test_const(const, env):
    os.environ.update(env)
    reload(const)
    if env:
        assert const.IFO == env['IFO']
        assert const.ifo == const.IFO.lower()
    else:
        assert const.IFO is None
        assert const.ifo is None
    assert const.site is None


@pytest.mark.parametrize('gps, default, epoch', [
    (1126259462, None, 'ER8'),
    (1187008882.43, None, 'O2'),
    (-1, 'test', 'test'),
])
def test_gps_epoch(gps, default, epoch):
    assert _const.gps_epoch(gps, default=default) == epoch


def test_gps_epoch_valueerror():
    with pytest.raises(ValueError):
        _const.gps_epoch(-1, default=None)


def test_latest_epoch(const):
    const.EPOCH = {
        'test1': (0, 10),
        'test2': (10, 20),
    }
    assert const.latest_epoch() == 'test2'



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for analysing ADC or DAC overflows
"""

import os
import re
from operator import attrgetter

import numpy

from gwdatafind import find_urls

from gwpy.time import to_gps
from gwpy.io.gwf import get_channel_names
from gwpy.timeseries import StateTimeSeries

from . import const
from .utils import natural_sort

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

_CHANNELS = {}


def find_overflows(timeseries, cumulative=True):
    """Find the times of overflows from an overflow counter

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        the input data from the cumulative overflow counter

    cumulative : `bool`, default: `True`
        whether the timeseries contains a cumulative overflow counter
        or an overflow state [0/1]

    Returns
    -------
    times : `numpy.ndarray`
        an array of GPS times (`~numpy.float64`) at which overflows
        were recorded
    """
    if cumulative:
        newoverflow = numpy.diff(
            (numpy.diff(timeseries.value) != 0).astype(int)) > 0
        return timeseries.times.value[2:][newoverflow]
    else:
        newoverflow = numpy.diff(timeseries.value) == 1
        return timeseries.times.value[1:][newoverflow]


def find_overflow_segments(timeseries, cumulative=True, round=False):
    """Find the segments during which the given counter indicated an overflow

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        the input data from the cumulative overflow counter

    cumulative : `bool`, default: `True`
        whether the timeseries contains a cumulative overflow counter
        or an overflow state [0/1]

    Returns
    -------
    overflows : `~gwpy.segments.SegmentList`
        a list of overflow segments
    """
    if cumulative:
        overflowing = (numpy.diff(timeseries) != 0)
        # rejig times after diff
        overflowing.x0 = timeseries.x0 + timeseries.dx
    else:
        overflowing = timeseries.astype(bool)
    return overflowing.view(StateTimeSeries).to_dqflag(
        name=timeseries.name, round=round)


def ligo_accum_overflow_channel(dcuid, ifo=None):
    """Returns the channel name for cumulative overflows for this DCUID

    Parameters
    ----------
    dcuid : `int`
        the DCUID for the relevant front-end model

    ifo : `str`
        the IFO prefix, defaults to the $IFO environment variable

    Returns
    -------
    channel : `str`
        the name of the cumulative overflow counter channel
    """
    ifo = ifo or const.IFO
    if ifo is None:
        raise ValueError("Cannot format channel without an IFO, "
                         "please specify")
    return '%s:FEC-%d_ACCUM_OVERFLOW' % (ifo, dcuid)


def ligo_model_overflow_channels(dcuid, ifo=None, frametype=None, gpstime=None,
                                 accum=True, nds=None):
    """Find the CDS overflow channel names for a given DCUID

    Parameters
    ----------
    dcuid : `int`
        the ID of the front-end controller to search

    ifo : `str`, optional
        the prefix of the interferometer to use

    frametype : `str`, optional
        the frametype to use, defaults to ``{ifo}_R``

    gpstime : `int`, optional
        the GPS time at which to search

    accum : `bool`, optional
        whether to retun the accumulated overflow channels (`True`) or not
        (`False`)

    nds : `str`, optional
        the ``'host:port'`` to use for accessing data via NDS, or `None`
        to use direct GWF file access

    Returns
    -------
    names : `list` of `str`
        the list of channel names found

    """
    ifo = ifo or const.IFO
    if ifo is None:
        raise ValueError("Cannot format channel without an IFO, "
                         "please specify")
    frametype = '{0}_R'.format(ifo)

    if gpstime is None:
        gpstime = int(to_gps('now')) - 1000

    if nds:
        allchannels = _ligo_model_overflow_channels_nds(dcuid, ifo, gpstime,
                                                        nds)
    else:
        allchannels = _ligo_model_overflow_channels_gwf(dcuid, ifo, frametype,
                                                        gpstime)

    if accum:
        regex = re.compile(r'%s:FEC-%d_(ADC|DAC)_OVERFLOW_ACC_\d+_\d+\Z'
                           % (ifo, dcuid))
    else:
        regex = re.compile(r'%s:FEC-%d_(ADC|DAC)_OVERFLOW_\d+_\d+\Z'
                           % (ifo, dcuid))
    return natural_sort(filter(regex.match, allchannels))


def _ligo_model_overflow_channels_nds(dcuid, ifo, gpstime, host):
    import nds2

    if host is True:
        try:
            host = os.getenv('NDSSERVER', '').split(',', 1)[0]
        except KeyError:
            raise ValueError("Cannot determine default NDSSERVER, please pass "
                             "nds=<host:port> or set NDSSERVER environment "
                             "variable")
    try:
        host, port = host.rsplit(':', 1)
    except ValueError:
        connection = nds2.connection(host)
    else:
        connection = nds2.connection(host, int(port))

    if connection.get_protocol() > 1:
        connection.set_epoch(gpstime, gpstime + 1)

    # NOTE: the `3` here is the channel type mask for 'ONLINE | RAW'
    return map(attrgetter('name'), connection.find_channels(
        '{ifo}:FEC-{dcuid}_*_OVERFLOW_*'.format(ifo=ifo, dcuid=dcuid), 3))


def _ligo_model_overflow_channels_gwf(dcuid, ifo, frametype, gpstime):
    try:
        framefile = find_urls(ifo[0], frametype, gpstime, gpstime)[0]
    except IndexError as e:
        e.args = ('No %s-%s frames found at GPS %d'
                  % (ifo[0], frametype, gpstime),)
        raise
    try:
        return _CHANNELS[framefile]
    except KeyError:
        _CHANNELS[framefile] = get_channel_names(framefile)
        return _CHANNELS[framefile]


def find_crossings(timeseries, threshold):
    """Find the times that a timeseries crosses a specific value

    Parameters
    ----------
    timeseries : `~gwpy.timeseries.TimeSeries`
        the input data to test against a threshold

    threshold : `float`
        function will analyze input timeseries and find times when data
        crosses this threshold

    Returns
    -------
    times : `numpy.ndarray`
        an array of GPS times (`~numpy.float64`) at which the input data
        crossed the threshold
    """
    if threshold >= 0:
        crossing_idx = numpy.nonzero(numpy.diff(
            (timeseries.value >= threshold).astype(int)
        ))[0] + 1
    else:
        crossing_idx = numpy.nonzero(numpy.diff(
            (timeseries.value > threshold).astype(int)
        ))[0] + 1

    return timeseries.times.value[crossing_idx]



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.io.html`
"""

import datetime
import os
import pytest
import shutil
import sys

from getpass import getuser
from MarkupPy import markup
from pygments import __version__ as pygments_version
from pytz import reference
from unittest import mock

from gwpy.segments import (Segment, DataQualityFlag)

from .. import html
from ..._version import get_versions
from ...utils import parse_html

from matplotlib import use
use('Agg')

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# global test objects

VERSION = get_versions()['version']
COMMIT = get_versions()['full-revisionid']

STYLESHEETS = '\n'.join([
    '<link href="{}" rel="stylesheet" media="all" />'.format(css)
    for css in html.CSS_FILES])
SCRIPTS = '\n'.join([
    '<script src="{}" type="text/javascript"></script>'.format(js)
    for js in html.JS_FILES])

NEW_BOOTSTRAP_PAGE = """<!DOCTYPE HTML>
<html lang="en">
<head>
<meta http-equiv="refresh" content="60" />
<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
<meta content="width=device-width, initial-scale=1.0" name="viewport" />
<base href="{base}" />
%s
%s
</head>
<body>
<div class="container">
</body>
</html>""" % (STYLESHEETS, SCRIPTS)

TEST_CONFIGURATION = """[section]
key = value"""

if pygments_version >= "2.14.0":
    pygments_space_span = '<span style="color: #bbbbbb"> </span>'
    pygments_output = (
        '\n'
        '<span style="color: #687822">key</span>'
        f'{pygments_space_span}'
        '<span style="color: #666666">=</span>'
        f'{pygments_space_span}'
        '<span style="color: #BA2121">value</span>'
    )
elif pygments_version >= "2.11.0":
    pygments_output = (
        '<span style="color: #bbbbbb"></span>\n'
        '<span style="color: #687822">key</span>'
        '<span style="color: #bbbbbb"> </span>'
        '<span style="color: #666666">=</span>'
        '<span style="color: #bbbbbb"> </span>'
        '<span style="color: #BA2121">value</span>'
        '<span style="color: #bbbbbb"></span>'
    )
    pygments_space_span = ' '
else:
    pygments_output = (
        '\n'
        '<span style="color: #7D9029">key</span> '
        '<span style="color: #666666">=</span> '
        '<span style="color: #BA2121">value</span>'
    )

ABOUT = f"""<div class="row">
<div class="col-md-12">
<h2>On the command-line</h2>
<p>This page was generated with the following command-line call:</p>
<div class="highlight" style="background: #f8f8f8"><pre style="line-height: 125%;"><span></span>${pygments_space_span}gwdetchar-scattering{pygments_space_span}-i{pygments_space_span}X1
</pre></div>

<p>The install path used was <code>{sys.prefix}</code>.</p>
<h2>Configuration files</h2>
<p>The following INI-format configuration file(s) were passed on the comand-line and are reproduced here in full:</p>
<div class="highlight" style="background: #f8f8f8"><pre style="line-height: 125%;"><span></span><span style="color: #008000; font-weight: bold">[section]</span>{pygments_output}
</pre></div>

<h2 class="mt-4">Environment</h2><table class="table table-sm table-hover table-responsive" id="package-table"><caption>Table of packages installed in the production environment</caption><thead><tr><th scope="col">Name</th><th scope="col">Version</th></tr></thead><tbody><tr><td>gwdetchar</td><td>1.2.3</td></tr><tr><td>gwpy</td><td>1.0.0</td></tr></tbody></table><button class="btn btn-outline-secondary btn-table mt-2" data-table-id="package-table" data-filename="package-table.csv">Export to CSV</button>
</div>
</div>"""  # noqa: E501

ABOUT_WITH_CONFIG_LIST = f"""<div class="row">
<div class="col-md-12">
<h2>On the command-line</h2>
<p>This page was generated with the following command-line call:</p>
<div class="highlight" style="background: #f8f8f8"><pre style="line-height: 125%;"><span></span>${pygments_space_span}gwdetchar-scattering{pygments_space_span}-i{pygments_space_span}X1
</pre></div>

<p>The install path used was <code>{sys.prefix}</code>.</p>
<h2>Configuration files</h2>
<p>The following INI-format configuration file(s) were passed on the comand-line and are reproduced here in full:</p>
<div id="accordion">
<div class="card mb-1 shadow-sm">
<div class="card-header">
<a class="collapsed card-link cis-link" href="#file0" data-toggle="collapse">test.ini</a>
</div>
<div id="file0" class="collapse" data-parent="#accordion">
<div class="card-body">
<div class="highlight" style="background: #f8f8f8"><pre style="line-height: 125%;"><span></span><span style="color: #008000; font-weight: bold">[section]</span>{pygments_output}
</pre></div>

</div>
</div>
</div>
</div>
<h2 class="mt-4">Environment</h2><table class="table table-sm table-hover table-responsive" id="package-table"><caption>Table of packages installed in the production environment</caption><thead><tr><th scope="col">Name</th><th scope="col">Version</th></tr></thead><tbody><tr><td>gwdetchar</td><td>1.2.3</td></tr><tr><td>gwpy</td><td>1.0.0</td></tr></tbody></table><button class="btn btn-outline-secondary btn-table mt-2" data-table-id="package-table" data-filename="package-table.csv">Export to CSV</button>
</div>
</div>"""  # noqa: E501

HTML_FOOTER = """<footer class="footer">
<div class="container">
<div class="row">
<div class="col-sm-3 icon-bar">
<a href="https://github.com/gwdetchar/gwdetchar/tree/%s" title="View gwdetchar-%s on GitHub" target="_blank"><i class="fas fa-code"></i></a>
<a href="https://github.com/gwdetchar/gwdetchar/issues" title="Open an issue ticket" target="_blank"><i class="fas fa-ticket-alt"></i></a>
<a href="about" title="How was this page generated?"><i class="fas fa-info-circle"></i></a>
<a href="external" title="View this page&quot;s external source"><i class="fas fa-external-link-alt"></i></a>
<a href="https://apod.nasa.gov/apod/astropix.html" title="Take a break from science" target="_blank"><i class="fas fa-heartbeat"></i></a>
</div>
<div class="col-sm-6">
<p>Created by {user} at {date}</p>
</div>
</div>
</div>
</footer>""" % (COMMIT, VERSION)  # noqa: E501

HTML_CLOSE = """</div>
%s
</body>
</html>""" % HTML_FOOTER

FLAG_CONTENT = """<div class="card border-warning mb-1 shadow-sm">
<div class="card-header text-white bg-warning">
<a class="collapsed card-link cis-link" href="#flag0" data-toggle="collapse">X1:TEST_FLAG</a>
</div>
<div id="flag0" class="collapse" data-parent="#accordion">
<div class="card-body">{plots}
{content}
</div>
</div>
</div>"""  # noqa: E501

FLAG_HTML = FLAG_CONTENT.format(content="""<pre># seg\tstart\tstop\tduration
0\t0\t66\t66.0
</pre>""", plots='')

FLAG_HTML_WITH_PLOTS = FLAG_CONTENT.format(
    content='<pre># seg\tstart\tstop\tduration\n0\t0\t66\t66.0\n</pre>',
    plots='\n<div class="row scaffold">\n<div class="col-sm-12">\n<a '
          'href="plots/X1-TEST_FLAG-0-66.png" id="a_X1-TEST_FLAG_66" '
          'title="Known (small) and active (large) analysis segments for '
          'X1:TEST_FLAG" class="fancybox" target="_blank" data-caption='
          '"Known (small) and active (large) analysis segments for '
          'X1:TEST_FLAG" data-fancybox="gallery" data-fancybox-group="images"'
          '>\n<img id="img_X1-TEST_FLAG_66" alt="X1-TEST_FLAG-0-66.png" '
          'class="img-fluid w-100" src="plots/X1-TEST_FLAG-0-66.png" />\n</a>'
          '\n</div>\n</div>')

FLAG_HTML_NO_SEGMENTS = FLAG_CONTENT.format(
    content='<p>No segments were found.</p>', plots='')

FLAG = DataQualityFlag(known=[(0, 66)], active=[(0, 66)], name='X1:TEST_FLAG')

OMEGA_SCAFFOLD = """<div class="card card-x1">
<div class="card-header pb-0">
<h5 class="card-title"><a class="cis-link" href="https://cis.ligo.org/channel/byname/X1:STRAIN" title="CIS entry for X1:STRAIN" target="_blank">X1:STRAIN</a></h5>
</div>
<div class="card-body">
<ul class="list-group">
<li class="list-group-item">
<div class="container">
<div class="row">
<h6>
<a href="./1126259462" class="text-dark">1126259462</a>
</h6>
</div>
<div class="row scaffold">
<div class="col-sm-4">
<a href="./1126259462/plots/X1-STRAIN-qscan_whitened-1.png" id="a_X1-STRAIN_1" title="X1-STRAIN-qscan_whitened-1.png" class="fancybox" target="_blank" data-caption="X1-STRAIN-qscan_whitened-1.png" data-fancybox="gallery" data-fancybox-group="images">
<img id="img_X1-STRAIN_1" alt="X1-STRAIN-qscan_whitened-1.png" class="img-fluid w-100 lazy" data-src="./1126259462/plots/X1-STRAIN-qscan_whitened-1.png" />
</a>
</div>
<div class="col-sm-4">
<a href="./1126259462/plots/X1-STRAIN-qscan_whitened-4.png" id="a_X1-STRAIN_4" title="X1-STRAIN-qscan_whitened-4.png" class="fancybox" target="_blank" data-caption="X1-STRAIN-qscan_whitened-4.png" data-fancybox="gallery" data-fancybox-group="images">
<img id="img_X1-STRAIN_4" alt="X1-STRAIN-qscan_whitened-4.png" class="img-fluid w-100 lazy" data-src="./1126259462/plots/X1-STRAIN-qscan_whitened-4.png" />
</a>
</div>
<div class="col-sm-4">
<a href="./1126259462/plots/X1-STRAIN-qscan_whitened-16.png" id="a_X1-STRAIN_16" title="X1-STRAIN-qscan_whitened-16.png" class="fancybox" target="_blank" data-caption="X1-STRAIN-qscan_whitened-16.png" data-fancybox="gallery" data-fancybox-group="images">
<img id="img_X1-STRAIN_16" alt="X1-STRAIN-qscan_whitened-16.png" class="img-fluid w-100 lazy" data-src="./1126259462/plots/X1-STRAIN-qscan_whitened-16.png" />
</a>
</div>
</div>
</div>
</li>
</ul>
</div>
</div>"""  # noqa: E501

CMDLINE = f"""<p>This page was generated with the following command-line call:</p>
<div class="highlight" style="background: #f8f8f8"><pre style="line-height: 125%;"><span></span>${pygments_space_span}gwdetchar-conlog{pygments_space_span}-i{pygments_space_span}X1
</pre></div>

<p>The install path used was <code>{sys.prefix}.</p>"""  # noqa: E501


# -- HTML unit tests ----------------------------------------------------------

def test_fancy_plot():
    # create a dummy FancyPlot instance
    test = html.FancyPlot('test.png')
    assert test.img == 'test.png'
    assert test.caption == 'test.png'

    # check that its properties are unchanged when the argument
    # to FancyPlot() is also a FancyPlot instance
    test = html.FancyPlot(test)
    assert test.img == 'test.png'
    assert test.caption == 'test.png'


def test_finalize_static_urls(tmpdir):
    base = str(tmpdir)
    static = os.path.join(base, 'static')
    (css, js) = html.finalize_static_urls(
        static, base, html.CSS_FILES, html.JS_FILES)
    assert set(css) == set(html.CSS_FILES)
    assert set(js) == set(html.JS_FILES)
    shutil.rmtree(str(tmpdir), ignore_errors=True)


def test_new_bootstrap_page():
    base = os.path.abspath(os.path.curdir)
    page = html.new_bootstrap_page(base=base, topbtn=False, refresh=True)
    assert parse_html(str(page)) == parse_html(
        NEW_BOOTSTRAP_PAGE.format(base=base))


def test_navbar():
    navbar = html.navbar(['test'], collapse=False)
    assert parse_html(navbar) == parse_html(
        '<nav class="navbar navbar-expand-md fixed-top shadow-sm">\n'
        '<div class="container-fluid">\n<div>\n'
        '<ul class="nav navbar-nav mr-auto">\n'
        '<li class="nav-item navbar-text">\n'
        'test\n</li>\n</ul>\n</div>\n</div>\n</nav>')


def test_dropdown():
    menu = html.dropdown('test', [])
    assert parse_html(str(menu)) == parse_html(
        '<a href="#" class="nav-link dropdown-toggle" role="button" '
        'data-toggle="dropdown">test</a>\n<div class="dropdown-menu '
        'dropdown-1-col shadow">\n</div>')

    menu = html.dropdown('test', ['test', '#'], active=0)
    assert parse_html(str(menu)) == parse_html(
        '<a href="#" class="nav-link dropdown-toggle" role="button" '
        'data-toggle="dropdown">test</a>\n<div class="dropdown-menu '
        'dropdown-1-col shadow">\ntest\n#\n</div>')

    menu = html.dropdown('test', ['test', '#'], active=[0, 1])
    assert parse_html(str(menu)) == parse_html(
        '<a href="#" class="nav-link dropdown-toggle" role="button" '
        'data-toggle="dropdown">test</a>\n<div class="dropdown-menu '
        'dropdown-1-col shadow">\ntest\n#\n</div>')

    menu = html.dropdown('<test />', ['test', '#'], active=[0, 1])
    assert parse_html(str(menu)) == parse_html(
        '<a href="#" class="nav-link dropdown-toggle" role="button" '
        'data-toggle="dropdown"><test /></a>\n<div class="dropdown-menu '
        'dropdown-1-col shadow">\ntest\n#\n</div>')


def test_dropdown_link():
    page = markup.page()
    html.dropdown_link(page, None)
    assert parse_html(str(page)) == parse_html(
        '<div class="dropdown-divider"></div>')

    page = markup.page()
    html.dropdown_link(page, 'test', active=True)
    assert parse_html(str(page)) == parse_html('test')

    page = markup.page()
    html.dropdown_link(page, 'test')
    assert parse_html(str(page)) == parse_html('test')


def test_get_brand():
    ((brand, help_), class_) = html.get_brand('H1', 'Test', 0, about='about')
    assert class_ == 'navbar fixed-top navbar-expand-md navbar-h1 shadow-sm'
    assert parse_html(brand) == parse_html(
        '<div class="navbar-brand border border-white rounded">H1 Test</div>')
    assert parse_html(help_) == parse_html(
        '<ul class="nav navbar-nav">\n<li class="nav-item dropdown">\n'
        '<a class="nav-link dropdown-toggle" href="#" role="button" '
        'data-toggle="dropdown">Links</a>\n<div class="dropdown-menu '
        'dropdown-menu-right shadow">\n<h6 class="dropdown-header">Internal'
        '</h6>\n<a href="about" class="dropdown-item">About this page</a>\n'
        '<div class="dropdown-divider"></div>\n<h6 class="dropdown-header">'
        'External</h6>\n<a href="https://ldas-jobs.ligo-wa.caltech.edu/'
        '~detchar/summary/day/19800106" class="dropdown-item" target="_blank">'
        'LHO Summary Pages</a>\n<a href="https://alog.ligo-wa.caltech.edu/'
        'aLOG" class="dropdown-item" target="_blank">LHO Logbook</a>\n'
        '</div>\n</li>\n</ul>')


@mock.patch(
    "gwdetchar.io.html.package_list",
    return_value=[
        {"name": "gwpy", "version": "1.0.0"},
        {"name": "gwdetchar", "version": "1.2.3"},
    ],
)
def test_about_this_page(package_list, tmpdir):
    outdir = str(tmpdir)
    config_file = os.path.join(outdir, 'test.ini')
    with open(config_file, 'w') as fobj:
        fobj.write(TEST_CONFIGURATION)
    testargs = ['/opt/bin/gwdetchar-scattering',
                '-i', 'X1']
    with mock.patch.object(sys, 'argv', testargs):
        # test with a single config file
        about = html.about_this_page(config_file)
        assert parse_html(about) == parse_html(ABOUT)
        # test with a list of config files
        about = html.about_this_page([config_file])
        assert parse_html(about) == parse_html(ABOUT_WITH_CONFIG_LIST)
    # clean up
    shutil.rmtree(outdir, ignore_errors=True)


def test_get_command_line():
    testargs = ['/opt/bin/gwdetchar-conlog',
                '--html-only',
                '-i', 'X1']
    with mock.patch.object(sys, 'argv', testargs):
        cmdline = str(html.get_command_line())
        assert (f'gwdetchar-conlog{pygments_space_span}-i'
                f'{pygments_space_span}X1') in cmdline
        assert not ('--html-only' in cmdline)
        assert 'The install path used was <code>{}</code>'.format(
            sys.prefix) in cmdline


@pytest.mark.parametrize('args, kwargs, result', [
    (('test.html', 'Test link'), {},
     '<a href="test.html" target="_blank">Test link</a>'),
    (('test.html', 'Test link'), {'class_': 'test-case'},
     '<a class="test-case" href="test.html" target="_blank">Test link</a>'),
])
def test_html_link(args, kwargs, result):
    h1 = parse_html(html.html_link(*args, **kwargs))
    h2 = parse_html(result)
    assert h1 == h2


def test_cis_link():
    h1 = parse_html(html.cis_link('X1:TEST-CHANNEL'))
    h2 = parse_html(
        '<a class="cis-link" href="https://cis.ligo.org/channel/byname/'
        'X1:TEST-CHANNEL" target="_blank" title="CIS entry for '
        'X1:TEST-CHANNEL">X1:TEST-CHANNEL</a>'
    )
    assert h1 == h2


def test_fancybox_img():
    img = html.FancyPlot('X1-TEST_AUX-test-4.png')
    out = html.fancybox_img(img)
    assert parse_html(out) == parse_html(
        '<a href="X1-TEST_AUX-test-4.png" id="a_X1-TEST_AUX_4" '
        'title="X1-TEST_AUX-test-4.png" class="fancybox" target="_blank" '
        'data-caption="X1-TEST_AUX-test-4.png" data-fancybox="gallery" '
        'data-fancybox-group="images">\n<img id="img_X1-TEST_AUX_4" '
        'alt="X1-TEST_AUX-test-4.png" class="img-fluid w-100" '
        'src="X1-TEST_AUX-test-4.png" />\n</a>')


def test_scaffold_plots():
    h1 = html.scaffold_plots([
        html.FancyPlot('X1-TEST_AUX-test-4.png'),
        html.FancyPlot('X1-TEST_AUX-test-16.png')], nperrow=2)
    assert parse_html(h1) == parse_html(
        '<div class="row scaffold">\n<div class="col-sm-6">\n'
        '<a href="X1-TEST_AUX-test-4.png" id="a_X1-TEST_AUX_4" '
        'title="X1-TEST_AUX-test-4.png" class="fancybox" target="_blank" '
        'data-caption="X1-TEST_AUX-test-4.png" data-fancybox="gallery" '
        'data-fancybox-group="images">\n'
        '<img id="img_X1-TEST_AUX_4" alt="X1-TEST_AUX-test-4.png" '
        'class="img-fluid w-100 lazy" data-src="X1-TEST_AUX-test-4.png" />\n'
        '</a>\n</div>\n<div class="col-sm-6">\n'
        '<a href="X1-TEST_AUX-test-16.png" id="a_X1-TEST_AUX_16" '
        'title="X1-TEST_AUX-test-16.png" class="fancybox" target="_blank" '
        'data-caption="X1-TEST_AUX-test-16.png" data-fancybox="gallery" '
        'data-fancybox-group="images">\n'
        '<img id="img_X1-TEST_AUX_16" alt="X1-TEST_AUX-test-16.png" '
        'class="img-fluid w-100 lazy" data-src="X1-TEST_AUX-test-16.png" />\n'
        '</a>\n</div>\n</div>')


def test_download_btn():
    page = html.download_btn([('test', 'test')])
    assert parse_html(page) == parse_html(
        '<div class="dropdown float-right d-none d-lg-block">\n'
        '<button type="button" class="btn btn-outline-secondary '
        'dropdown-toggle" data-toggle="dropdown" aria-expanded="false" '
        'aria-haspopup="true">Download summary</button>\n<div '
        'class="dropdown-menu dropdown-menu-right shadow">\n<a href="test" '
        'download="test" class="dropdown-item">test</a>\n</div>\n</div>')


def test_parameter_table():
    page = html.parameter_table([('test', 'test')],
                                start=0, end=1, flag='X1:TEST')
    assert '<th scope="row">Start time (UTC)</th>' in page
    assert '<td>1980-01-06 00:00:00 (0)</td>' in page
    assert '<th scope="row">End time (UTC)</th>' in page
    assert '<td>1980-01-06 00:00:01 (1)</td>' in page
    assert '<th scope="row">State flag</th>' in page
    assert '<td><code>X1:TEST</code></td>' in page
    assert '<th scope="row">test</th>' in page
    assert '<td>test</td>' in page


def test_alert():
    page = html.alert('test')
    assert parse_html(page) == parse_html(
        '<div class="alert alert-info alert-dismissible fade show text-justify'
        ' shadow-sm">\n<button type="button" class="close" data-dismiss='
        '"alert" aria-label="Close">\n<span aria-hidden="true">&times;</span>'
        '\n</button>\ntest\n</div>')


def test_alert_with_list():
    page = html.alert(['test'])
    assert parse_html(page) == parse_html(
        '<div class="alert alert-info alert-dismissible fade show text-justify'
        ' shadow-sm">\n<button type="button" class="close" data-dismiss='
        '"alert" aria-label="Close">\n<span aria-hidden="true">&times;</span>'
        '\n</button>\n<p>test</p>\n</div>')


def test_table():
    headers = ['Test']
    data = [['test']]
    caption = 'This is a test table.'
    page = html.table(headers=headers, data=data, caption=caption, id='test')
    assert parse_html(page) == parse_html(
        '<table class="table table-sm table-hover" id="test"><caption>'
        'This is a test table.</caption><thead><tr><th scope="col">Test'
        '</th></tr></thead><tbody><tr><td>test</td></tr></tbody></table>'
        '<button class="btn btn-outline-secondary btn-table mt-2" '
        'data-table-id="test" data-filename="test.csv">Export to CSV</button>')


def test_write_flag_html():
    page = html.write_flag_html(FLAG)
    assert parse_html(str(page)) == parse_html(FLAG_HTML)

    page2 = html.write_flag_html(
        DataQualityFlag(known=[], active=[], name='X1:TEST_FLAG'))
    assert parse_html(str(page2)) == parse_html(FLAG_HTML_NO_SEGMENTS)


def test_write_flag_html_with_plots(tmpdir):
    tmpdir.mkdir('plots')
    os.chdir(str(tmpdir))
    page = html.write_flag_html(FLAG, span=Segment(0, 66), plotdir='plots')
    assert parse_html(str(page)) == parse_html(FLAG_HTML_WITH_PLOTS)
    shutil.rmtree(str(tmpdir), ignore_errors=True)


def test_scaffold_omega_scans():
    times = [1126259462]
    channel = 'X1:STRAIN'
    page = html.scaffold_omega_scans(times, channel)
    assert parse_html(page) == parse_html(OMEGA_SCAFFOLD)


def test_write_footer():
    now = datetime.datetime.now()
    tz = reference.LocalTimezone().tzname(now)
    date = now.strftime('%H:%M {} on %d %B %Y'.format(tz))
    out = html.write_footer(about='about', external='external')
    assert parse_html(str(out)) == parse_html(
        HTML_FOOTER.format(user=getuser(), date=date))
    with pytest.raises(ValueError) as exc:
        html.write_footer(link='test')
    assert 'argument must be either None or a tuple' in str(exc.value)


def test_close_page(tmpdir):
    target = os.path.join(str(tmpdir), 'test.html')
    now = datetime.datetime.now()
    tz = reference.LocalTimezone().tzname(now)
    date = now.strftime('%H:%M {} on %d %B %Y'.format(tz))
    page = html.close_page(html.markup.page(), target,
                           about='about', external='external')
    assert parse_html(str(page)) == parse_html(
        HTML_CLOSE.format(user=getuser(), date=str(date)))
    assert os.path.isfile(target)
    with open(target, 'r') as fp:
        assert fp.read() == str(page)
    shutil.rmtree(target, ignore_errors=True)


@mock.patch("{}.Path.is_dir".format(html.Path.__module__))
@mock.patch("subprocess.check_output", return_value="{\"key\": 0}")
@pytest.mark.parametrize("isdir, cmd", [
    pytest.param(
        False,
        "{} -m pip list installed --format json".format(sys.executable),
        id="pip",
    ),
    pytest.param(
        True,
        "conda list --prefix {} --json".format(sys.prefix),
        id="conda",
    ),
])
def test_package_list(check_output, is_dir, isdir, cmd):
    is_dir.return_value = isdir
    assert html.package_list() == {"key": 0}
    check_output.assert_called_with(cmd.split())


@mock.patch(
    "gwdetchar.io.html.package_list",
    return_value=[
        {"name": "gwpy", "version": "1.0.0"},
        {"name": "gwdetchar", "version": "1.2.3"},
    ],
)
def test_package_table(package_list):
    assert parse_html(
        html.package_table(class_="test", caption="Test"),
    ) == parse_html(
        '<h2 class="mt-4">Environment</h2>'
        '<table class="test" id="package-table"><caption>'
        'Test</caption><thead><tr><th scope="col">Name</th><th scope="col">'
        'Version</th></tr></thead><tbody><tr><td>gwdetchar</td><td>1.2.3</td>'
        '</tr><tr><td>gwpy</td><td>1.0.0</td></tr></tbody></table><button '
        'class="btn btn-outline-secondary btn-table mt-2" data-table-id='
        '"package-table" data-filename="package-table.csv">Export to CSV'
        '</button>'
    )



# coding=utf-8
# Copyright (C) LIGO Scientific Collaboration (2015-)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Find channels clipping their software saturation limits
"""

import gwdatafind
import os
import sys

from MarkupPy import markup

from gwpy.io.cache import sieve as sieve_cache
from gwpy.io.gwf import get_channel_names
from gwpy.segments import (Segment, SegmentList,
                           DataQualityFlag, DataQualityDict)

from . import core
from .. import (cli, const)
from ..io import html as htmlio

from matplotlib import use
use('Agg')

__author__ = 'Dan Hoak <daniel.hoak@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# set up logger
PROG = ('python -m gwdetchar.saturation' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = cli.logger(name=PROG.split('python -m ').pop())


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = cli.create_parser(
        prog=PROG,
        description=__doc__,
    )

    # required arguments
    cli.add_gps_start_stop_arguments(parser)
    cli.add_ifo_option(parser)

    # optional arguments
    cli.add_frametype_option(
        parser,
        required=const.IFO is None,
        default=const.IFO is not None and '%s_R' % const.IFO,
    )
    cli.add_nproc_option(parser)
    parser.add_argument(
        '-c',
        '--channels',
        help='file containing columnar list of channels to process, '
             'default is to find all relevant channels from frames',
    )
    parser.add_argument(
        '-s',
        '--skip',
        nargs='*',
        default=[],
        help='skip channels matching this string',
    )
    parser.add_argument(
        '-g',
        '--group-size',
        default=1024,
        type=int,
        help='number of channels to process in a single batch, '
             'default: %(default)s',
    )
    parser.add_argument(
        '-a',
        '--state-flag',
        metavar='FLAG',
        help='restrict search to times when FLAG was active',
    )
    parser.add_argument(
        '-p',
        '--pad-state-end',
        metavar='PAD',
        default=0,
        type=float,
        help='pad state segments inwards from the end '
             'by PAD segments, default: %(default)s',
    )
    parser.add_argument(
        '-m',
        '--html',
        type=os.path.abspath,
        help='path to write html output',
    )
    parser.add_argument(
        '-v',
        '--plot',
        action='store_true',
        default=False,
        help='make plots of all saturations, default: %(default)s',
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the software saturation command-line interface
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    # get IFO
    ifo = args.ifo.upper()
    site = ifo[0]
    frametype = args.frametype or '%s_R' % ifo

    # let's go
    LOGGER.info('{} Software saturations {}-{}'.format(
        args.ifo, int(args.gpsstart), int(args.gpsend)))

    # get segments
    span = Segment(args.gpsstart, args.gpsend)
    if args.state_flag:
        state = DataQualityFlag.query(args.state_flag, int(args.gpsstart),
                                      int(args.gpsend),
                                      url=const.DEFAULT_SEGMENT_SERVER)
        for i, seg in enumerate(state.active):
            state.active[i] = type(seg)(seg[0], seg[1]-args.pad_state_end)
        segs = state.active.coalesce()
        LOGGER.debug("Recovered %d seconds of time for %s"
                     % (abs(segs), args.state_flag))
    else:
        segs = SegmentList([Segment(args.gpsstart, args.gpsend)])

    # find frames
    cache = gwdatafind.find_urls(
        site, frametype, int(args.gpsstart), int(args.gpsend))

    # find channels
    if not os.getenv('LIGO_DATAFIND_SERVER'):
        raise RuntimeError("No LIGO_DATAFIND_SERVER variable set, don't know "
                           "how to discover channels")
    else:
        LOGGER.debug("Identifying channels in frame files")
        if len(cache) == 0:
            raise RuntimeError(
                "No frames recovered for %s in interval [%s, %s)" %
                (frametype, int(args.gpsstart),
                 int(args.gpsend)))
        allchannels = get_channel_names(cache[0])
        LOGGER.debug("   Found %d channels" % len(allchannels))
        sys.stdout.flush()
        channels = core.find_limit_channels(allchannels, skip=args.skip)
        LOGGER.info(
            "   Parsed %d channels with '_LIMIT' and '_LIMEN' or '_SWSTAT'"
            % sum(map(len, channels)))

    # -- read channels and check limits -------------

    saturations = DataQualityDict()
    bad = set()

    # TODO: use multiprocessing to separate channel list into discrete chunks
    #       should give a factor of X for X processes

    # check limens
    for suffix, clist in zip(['LIMEN', 'SWSTAT'], channels):
        nchans = len(clist)
        # group channels in sets for batch processing
        #     min of <number of channels>, user group size (sensible number),
        #     and 512 Mb of RAM for single-precision EPICS
        try:
            dur = max([float(abs(s)) for s in segs])
        except ValueError:
            ngroup = args.group_size
        else:
            ngroup = int(
                min(nchans, args.group_size, 2 * 1024**3 / 4. / 16. / dur))
        LOGGER.info('Processing %s channels in groups of %d' % (
            suffix, ngroup))
        sys.stdout.flush()
        sets = core.grouper(clist, ngroup)
        for i, cset in enumerate(sets):
            # remove empty entries use to pad the list to 8 elements
            cset = list(cset)
            while cset[-1] is None:
                cset.pop(-1)
            for seg in segs:
                cache2 = sieve_cache(cache, segment=seg)
                if not len(cache2):
                    continue
                saturated = core.is_saturated(
                    cset, cache2, seg[0], seg[1], indicator=suffix,
                    nproc=args.nproc)
                for new in saturated:
                    try:
                        saturations[new.name] += new
                    except KeyError:
                        saturations[new.name] = new
            for j, c in enumerate(cset):
                try:
                    sat = saturations[c]
                except KeyError:
                    LOGGER.debug('%40s:      SKIP      [%d/%d]'
                                 % (c, i*ngroup + j + 1, nchans))
                else:
                    if abs(sat.active):
                        LOGGER.debug('%40s: ---- FAIL ---- [%d/%d]'
                                     % (c, i*ngroup + j + 1, nchans))
                        for seg in sat.active:
                            LOGGER.debug(" " * 42 + str(seg))
                        bad.add(c)
                    else:
                        LOGGER.debug('%40s:      PASS      [%d/%d]'
                                     % (c, i*ngroup + j + 1, nchans))
                sys.stdout.flush()

    # -- log results and exit -----------------------

    if len(bad):
        LOGGER.info("Saturations were found for all of the following:\n\n")
        for c in bad:
            print(c)
        print('\n\n')
    else:
        LOGGER.info("No software saturations were found in any channels")

    # write segments to file
    outfile = ('%s-SOFTWARE_SATURATIONS-%d-%d.h5'
               % (ifo, int(args.gpsstart),
                  int(args.gpsend) - int(args.gpsstart)))
    LOGGER.info("Writing saturation segments to %s" % outfile)
    saturations.write(outfile, path="segments", overwrite=True)

    if args.html:
        # get base path
        base = os.path.dirname(args.html)
        os.chdir(base)
        if args.plot:
            args.plot = os.path.curdir
        segfile = os.path.relpath(outfile, os.path.dirname(args.html))
        if os.path.basename(args.html) == 'index.html':
            links = [
                '%d-%d' % (int(args.gpsstart), int(args.gpsend)),
                ('Parameters', '#parameters'),
                ('Segments', [('Software saturations',
                               '#software-saturations')]),
                ('Results', '#results'),
            ]
            if args.state_flag:
                links[2][1].insert(0, ('State flag', '#state-flag'))
            (brand, class_) = htmlio.get_brand(ifo, 'Saturations',
                                               args.gpsstart)
            navbar = htmlio.navbar(links, class_=class_, brand=brand)
            page = htmlio.new_bootstrap_page(
                navbar=navbar, title='%s Saturations | %d-%d' % (
                    ifo, int(args.gpsstart), int(args.gpsend)))
        else:
            page = markup.page()
            page.div(class_='container')
        # -- header
        page.div(class_='pb-2 mt-3 mb-2 border-bottom')
        page.h1('%s Software Saturations: %d-%d'
                % (ifo, int(args.gpsstart), int(args.gpsend)))
        page.div.close()
        # -- paramters
        content = [
            ('State end padding', args.pad_state_end),
            ('Skip', ', '.join(map(repr, args.skip)))]
        page.h2('Parameters', class_='mt-4 mb-4', id_='parameters')
        page.div(class_='row')
        page.div(class_='col-md-9 col-sm-12')
        page.add(htmlio.parameter_table(
            content, start=args.gpsstart, end=args.gpsend,
            flag=args.state_flag))
        page.div.close()  # col-md-9 col-sm-12
        page.div(class_='col-md-3 col-sm-12')
        page.add(htmlio.download_btn(
            [('Segments (HDF)', segfile)],
            btnclass='btn btn-%s dropdown-toggle' % ifo.lower(),
        ))
        page.div.close()  # col-md-9 col-sm-12
        page.div.close()  # row
        page.h5('Command-line:')
        page.add(htmlio.get_command_line(about=False, prog=PROG))
        # -- segments
        page.h2('Segments', class_='mt-4', id_='segments')
        msg = ("This analysis searched {0} filter bank readback channels for "
               "time periods during which their OUTPUT value matched or "
               "exceeded the LIMIT value set in software. Signals that "
               "achieve saturation are shown below, and saturation segments "
               "are available by expanding a given panel.").format(
                   sum(map(len, channels)))
        page.add(htmlio.alert(msg, context=ifo.lower()))
        # record state segments
        if args.state_flag:
            page.h3('State flag', class_='mt-3', id_='state-flag')
            page.div(id_='accordion1')
            page.add(htmlio.write_flag_html(
                state, span, 'state', parent='accordion1', context='success',
                plotdir=args.plot, facecolor=(0.2, 0.8, 0.2),
                edgecolor='darkgreen', known={
                    'facecolor': 'red',
                    'edgecolor': 'darkred',
                    'height': 0.4},
            ))
            page.div.close()
        # record saturation segments
        if len(bad):
            page.h3('Software saturations', class_='mt-3',
                    id_='software-saturations')
            page.div(id_='accordion2')
            for i, (c, flag) in enumerate(saturations.items()):
                if abs(flag.active) > 0:
                    title = '%s [%d]' % (flag.name, len(flag.active))
                    page.add(htmlio.write_flag_html(
                        flag, span=span, id=i, parent='accordion2',
                        title=title, plotdir=args.plot))
            page.div.close()
        else:
            page.add(htmlio.alert('No software saturations were found in this '
                                  'analysis', context=ifo.lower(),
                                  dismiss=False))
        # -- results table
        page.h2('Results summary', class_='mt-4', id_='results')
        page.add(htmlio.alert('All channels for which the LIMIT setting was '
                              'active are shown below.', context=ifo.lower()))
        page.table(class_='table table-striped table-hover')
        # write table header
        page.thead()
        page.tr()
        for header in ['Channel', 'Result', 'Num. saturations']:
            page.th(header)
        page.thead.close()
        # write body
        page.tbody()
        for c, seglist in saturations.items():
            passed = abs(seglist.active) == 0
            if passed:
                page.tr()
            else:
                page.tr(class_='table-warning')
            page.td(c)
            page.td(passed and 'Pass' or 'Fail')
            page.td(len(seglist.active))
            page.tr.close()
        page.tbody.close()
        page.table.close()
        # close and write
        htmlio.close_page(page, args.html)


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()



# coding=utf-8
# Copyright (C) Duncan Macleod (2018)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests of :mod:`gwdetchar.cds`
"""

import pytest

from io import BytesIO
from unittest import mock

from .. import cds

RTDCUID_CONTENT = """
<div class="listing">
  <code>1 x1model1</code>
  <code>2 x2model2</code>
  <code>nonint x2model3</code>
  <code>3 x3model3 test</code>
</div>
"""

ADCLIST_CONTENT = """
<div class="listing">
  <code>0 0 A X</code>
  <code>0 1 B X</code>
  <code>1 0 C X </code>
  <code>A B C D</code>
  <code>1	1	D	X</code>
</div>
"""


def mock_request(output):
    if isinstance(output, str):
        output = output.encode('utf-8')
    return mock.patch('gwdetchar.cds.request.urlopen',
                      return_value=BytesIO(output))


@mock_request(RTDCUID_CONTENT)
def test_get_dcuid_map(_):
    dcuidmap = cds.get_dcuid_map('X1')
    assert dcuidmap == {
        1: 'x1model1',
        2: 'x2model2',
    }


@mock_request(ADCLIST_CONTENT)
def test_get_adclist(_):
    adclist = cds.get_adclist('X1', 'x1model1')
    assert adclist == {
        (0, 0): 'A',
        (0, 1): 'B',
        (1, 0): 'C',
        (1, 1): 'D',
    }


@mock_request(RTDCUID_CONTENT)
def test_model_name_from_dcuid(mockr):
    cds.DCUID_MAP = {}
    assert cds.model_name_from_dcuid('X1', 1) == 'x1model1'
    assert cds.model_name_from_dcuid('X1', 2) == 'x2model2'
    assert mockr.call_count == 1
    with pytest.raises(KeyError):
        assert cds.model_name_from_dcuid('X1', 9) == 1


@mock_request(RTDCUID_CONTENT)
def test_dcuid_from_model_name(mockr):
    cds.DCUID_MAP = {}
    assert cds.dcuid_from_model_name('X1', 'x1model1') == 1
    assert cds.dcuid_from_model_name('X1', 'x2model2') == 2
    assert mockr.call_count == 1
    with pytest.raises(KeyError):
        assert cds.dcuid_from_model_name('X1', 'x1model9') == 1


@mock_request(ADCLIST_CONTENT)
def test_get_adc_channel(mockr):
    cds.ADC_MAP = {}
    assert cds.get_adc_channel('X1', 'x1model1', 0, 0) == 'A'
    assert cds.get_adc_channel('X1', 'x1model1', 0, 1) == 'B'
    assert mockr.call_count == 1
    with pytest.raises(KeyError):
        assert cds.get_adc_channel('X1', 'x1model1', 9, 9)


@mock_request(ADCLIST_CONTENT)
@mock.patch(
    'gwdetchar.cds.model_name_from_dcuid',
    return_value='x1model1',
)
def test_get_real_channel(mockr, mockadc):
    assert cds.get_real_channel('X1:TEST-1_ADC_OVERFLOW_ACC_0_0') == 'A'
    assert cds.get_real_channel('X1:TEST-1_ADC_OVERFLOW_ACC_0_1') == 'B'
    assert mockr.call_count == 2
    with pytest.raises(ValueError):
        assert cds.get_real_channel('X1:TEST-1_TEST_OVERFLOW_ACC_0_0')



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the GW DetChar python package.
#
# gwdetchar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# gwdetchar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwdetchar.utils`
"""

import pytest

import numpy

from gwpy.segments import (
    DataQualityFlag,
    DataQualityDict,
)
from gwpy.table import EventTable
from gwpy.testing.utils import assert_table_equal

from .. import utils


@pytest.mark.parametrize('in_, out', [
    ([1, 2, 3], [1, 2, 3]),
    ([1, 10, 2, 3], [1, 2, 3, 10]),
    (['1', '10', '2', '3'], ['1', '2', '3', '10']),
    (['a', 'b', 'd', 'c'], ['a', 'b', 'c', 'd']),
])
def test_natural_sort(in_, out):
    assert utils.natural_sort(in_) == out


def test_table_from_segments():
    segs = DataQualityDict()
    segs["test1"] = DataQualityFlag(
        active=[(0, 1), (3, 4)],
    )
    segs["test2"] = DataQualityFlag(
        active=[(5, 6)],
    )
    assert_table_equal(
        utils.table_from_segments(segs),
        EventTable(
            rows=[(0., 100., 0., 1., 10., "test1"),
                  (3., 100., 3., 4., 10., "test1"),
                  (5., 100., 5., 6., 10., "test2")],
            names=("time", "frequency", "start_time", "end_time",
                   "snr", "channel"),
        ),
    )
    assert_table_equal(
        utils.table_from_segments(
            segs,
            frequency=1234.,
            snr=-4.,
            sngl_burst=True),
        EventTable(
            rows=[(0., 1234., -4., "test1"),
                  (3., 1234., -4., "test1"),
                  (5., 1234., -4., "test2")],
            names=("peak", "peak_frequency", "snr", "channel"),
        ),
    )


def test_table_from_segments_empty():
    segs = DataQualityDict()
    segs['test'] = DataQualityFlag(
       active=[]
    )
    assert_table_equal(
        utils.table_from_segments(segs),
        EventTable(
            names=("time", "frequency", "start_time", "end_time",
                   "snr", "channel")
        )
    )


def test_table_from_times():
    times = numpy.array(range(10), dtype=float)
    assert_table_equal(
        utils.table_from_times(times),
        EventTable([times, [100.] * 10, [10.] * 10],
                   names=("time", "frequency", "snr")),
    )
